##############################################################################
#
# Copyright 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Python BTree implementation
"""

from persistent import Persistent

from ._compat import compare
from .Interfaces import BTreesConflictError


_marker = object()


class _Base(Persistent):

    __slots__ = ()
    # This is used to allocate storage for the keys.
    # It's probably here so that we could, for example, use
    # an ``array.array`` for native types. But nothing actually does
    # that, everything is stored boxed.
    # TODO: Figure out why not.
    _key_type = list

    def __init__(self, items=None):
        self.clear()
        if items:
            self.update(items)

    try:
        # Detect the presence of the C extensions.
        # If they're NOT around, we don't need to do any of the
        # special pickle support to make Python versions look like
        # C---we just rename the classes. By not defining these methods,
        # we can (theoretically) avoid a bit of a slowdown.
        # If the C extensions are around, we do need these methods, but
        # these classes are unlikely to be used in production anyway.
        __import__('BTrees._OOBTree')
    # ModuleNotFoundError not possible here, a real ImportError is raised:
    except ImportError:  # pragma: no cover
        pass
    else:
        def __reduce__(self):
            # Swap out the type constructor for the C version, if present.
            func, typ_gna, state = Persistent.__reduce__(self)
            # We ignore the returned type altogether in favor of
            # our calculated class (which allows subclasses but replaces our
            # exact type with the C equivalent)
            typ = self.__class__
            gna = typ_gna[1:]
            return (func, (typ,) + gna, state)

        @property
        def __class__(self):
            type_self = type(self)
            return (
                type_self._BTree_reduce_as
                if type_self._BTree_reduce_up_bound is type_self
                else type_self
            )

        @property
        def _BTree_reduce_as(self):
            # Return the pickle replacement class for this object.
            # If the C extensions are available, this will be the
            # C type (setup by _fix_pickle), otherwise it will be the real
            # type of this object.
            # This implementation is replaced by _fix_pickle and exists for
            # testing purposes.
            return type(self)  # pragma: no cover

        _BTree_reduce_up_bound = _BTree_reduce_as


class _ArithmeticMixin:

    def __sub__(self, other):
        return difference(self.__class__, self, other)

    def __rsub__(self, other):
        return difference(self._set_type, type(self)(other), self)

    def __or__(self, other):
        return union(self._set_type, self, other)

    __ror__ = __or__

    def __and__(self, other):
        return intersection(self._set_type, self, other)

    __rand__ = __and__

    def __xor__(self, other):
        return (self - other) | (other - self)

    __rxor__ = __xor__


class _BucketBase(_ArithmeticMixin, _Base):

    __slots__ = ('_keys', '_next', '_to_key')

    def clear(self):
        self._keys = self._key_type()
        self._next = None

    def __len__(self):
        return len(self._keys)

    @property
    def size(self):
        return len(self._keys)

    def _deleteNextBucket(self):
        next = self._next
        if next is not None:
            self._next = next._next

    def _search(self, key):
        # Return non-negative index on success
        # return -(insertion_index + 1) on fail
        low = 0
        keys = self._keys
        high = len(keys)
        while low < high:
            i = (low + high) // 2
            k = keys[i]
            if k is key or k == key:
                return i

            if compare(k, key) < 0:
                low = i + 1
            else:
                high = i
        return -1 - low

    def minKey(self, key=_marker):
        if key is _marker or key is None:
            return self._keys[0]
        key = self._to_key(key)
        index = self._search(key)
        if index >= 0:
            return key
        index = -index - 1
        if index < len(self._keys):
            return self._keys[index]
        else:
            raise ValueError("no key satisfies the conditions")

    def maxKey(self, key=_marker):
        if key is _marker or key is None:
            return self._keys[-1]
        key = self._to_key(key)
        index = self._search(key)
        if index >= 0:
            return key
        else:
            index = -index - 1
            if index:
                return self._keys[index - 1]
            else:
                raise ValueError("no key satisfies the conditions")

    def _range(self, min=_marker, max=_marker,
               excludemin=False, excludemax=False):
        if min is _marker or min is None:
            start = 0
            if excludemin:
                start = 1
        else:
            min = self._to_key(min)
            start = self._search(min)
            if start >= 0:
                if excludemin:
                    start += 1
            else:
                start = -start - 1
        if max is _marker or max is None:
            end = len(self._keys)
            if excludemax:
                end -= 1
        else:
            max = self._to_key(max)
            end = self._search(max)
            if end >= 0:
                if not excludemax:
                    end += 1
            else:
                end = -end - 1

        return start, end

    def keys(self, *args, **kw):
        start, end = self._range(*args, **kw)
        return self._keys[start:end]

    def iterkeys(self, *args, **kw):
        if not (args or kw):
            return iter(self._keys)
        keys = self._keys
        return (keys[i] for i in range(*self._range(*args, **kw)))

    def __iter__(self):
        return iter(self._keys)

    def __contains__(self, key):
        try:
            tree_key = self._to_key(key)
        except TypeError:
            # Can't convert the key, so can't possibly be in the tree
            return False

        return (self._search(tree_key) >= 0)

    has_key = __contains__

    def _repr_helper(self, items):
        type_self = type(self)
        mod = type_self.__module__
        name = type_self.__name__
        name = name[:-2] if name.endswith("Py") else name
        return f"{mod}.{name}({items!r})"


class _SetIteration:

    __slots__ = (
        'to_iterate',
        'useValues',
        '_iter',
        'active',
        'position',
        'key',
        'value',
    )

    def __init__(self, to_iterate, useValues=False, default=None, sort=False):
        if to_iterate is None:
            to_iterate = ()
        self.to_iterate = to_iterate
        if sort:
            # Sorting is required for arbitrary iterables in the
            # set functions like difference/union/intersection
            assert not useValues
            if not isinstance(to_iterate, _Base):
                # We know _Base (Set, Bucket, Tree, TreeSet) will all iterate
                # in sorted order. Other than that, we have no guarantee.
                self.to_iterate = to_iterate = sorted(self.to_iterate)

        if useValues:
            try:
                itmeth = to_iterate.iteritems
            except AttributeError:
                if isinstance(to_iterate, dict):
                    itmeth = to_iterate.items().__iter__
                else:
                    itmeth = to_iterate.__iter__
                    useValues = False
            else:
                self.value = None
        else:
            itmeth = to_iterate.__iter__

        self.useValues = useValues
        self._iter = itmeth()
        self.active = True
        self.position = 0
        self.key = _marker
        self.value = default
        self.advance()

    def advance(self):
        try:
            if self.useValues:
                self.key, self.value = next(self._iter)
            else:
                self.key = next(self._iter)
            self.position += 1
        except StopIteration:
            self.active = False
            self.position = -1

        return self


class _MutableMappingMixin:
    # Methods defined in collections.abc.MutableMapping that
    # Bucket and Tree should both implement and can implement
    # the same. We don't want to extend that class though,
    # as the C version cannot.
    def popitem(self):
        """
        D.popitem() -> (k, v), remove and return some (key, value) pair
        as a 2-tuple; but raise KeyError if D is empty.
        """
        try:
            key = next(iter(self))
        except StopIteration:
            raise KeyError
        value = self[key]
        del self[key]
        return key, value


class Bucket(_MutableMappingMixin, _BucketBase):

    __slots__ = ()
    _value_type = list
    VALUE_SAME_CHECK = False

    def _to_value(self, x):
        return x

    def setdefault(self, key, value):
        key, value = self._to_key(key), self._to_value(value)
        status, value = self._set(key, value, True)
        return value

    def pop(self, key, default=_marker):
        try:
            status, value = self._del(self._to_key(key))
        except KeyError:
            if default is _marker:
                raise
            return default
        else:
            return value

    def update(self, items):
        if hasattr(items, 'iteritems'):
            items = items.iteritems()
        elif hasattr(items, 'items'):
            items = items.items()

        _si = self.__setitem__
        try:
            for key, value in items:
                _si(key, value)
        except ValueError:
            raise TypeError('items must be a sequence of 2-tuples')

    def __setitem__(self, key, value):
        self._set(self._to_key(key), self._to_value(value))

    def __delitem__(self, key):
        self._del(self._to_key(key))

    def clear(self):
        _BucketBase.clear(self)
        self._values = self._value_type()

    def get(self, key, default=None):
        try:
            key = self._to_key(key)
        except TypeError:
            # Can't convert, cannot possibly be present.
            return default
        index = self._search(key)
        if index < 0:
            return default
        return self._values[index]

    def __getitem__(self, key):
        try:
            tree_key = self._to_key(key)
        except TypeError:
            # Can't convert, so cannot possibly be present.
            raise KeyError(key)
        index = self._search(tree_key)
        if index < 0:
            raise KeyError(key)
        return self._values[index]

    def _set(self, key, value, ifunset=False):
        """Set a value

        Return: status, value

        Status is:
              None if no change
                  0 if change, but not size change
                  1 if change and size change
        """
        index = self._search(key)
        if index >= 0:
            if (
                ifunset or
                self.VALUE_SAME_CHECK and value == self._values[index]
            ):
                return None, self._values[index]
            self._p_changed = True
            self._values[index] = value
            return 0, value
        else:
            self._p_changed = True
            index = -index - 1
            self._keys.insert(index, key)
            self._values.insert(index, value)
            return 1, value

    def _del(self, key):
        index = self._search(key)
        if index >= 0:
            self._p_changed = True
            del self._keys[index]
            return 0, self._values.pop(index)
        raise KeyError(key)

    def _split(self, index=-1):
        if index < 0 or index >= len(self._keys):
            index = len(self._keys) // 2
        new_instance = type(self)()
        new_instance._keys = self._keys[index:]
        new_instance._values = self._values[index:]
        del self._keys[index:]
        del self._values[index:]
        new_instance._next = self._next
        self._next = new_instance
        return new_instance

    def values(self, *args, **kw):
        start, end = self._range(*args, **kw)
        return self._values[start:end]

    def itervalues(self, *args, **kw):
        values = self._values
        return (values[i] for i in range(*self._range(*args, **kw)))

    def items(self, *args, **kw):
        keys = self._keys
        values = self._values
        return [
            (keys[i], values[i])
            for i in range(*self._range(*args, **kw))
        ]

    def iteritems(self, *args, **kw):
        keys = self._keys
        values = self._values
        return (
            (keys[i], values[i])
            for i in range(*self._range(*args, **kw))
        )

    def __getstate__(self):
        keys = self._keys
        values = self._values
        data = []
        for i in range(len(keys)):
            data.append(keys[i])
            data.append(values[i])
        data = tuple(data)

        if self._next is not None:
            return data, self._next
        return (data, )

    def __setstate__(self, state):
        if not isinstance(state[0], tuple):
            raise TypeError("tuple required for first state element")

        self.clear()
        if len(state) == 2:
            state, self._next = state
        else:
            self._next = None
            state = state[0]

        keys = self._keys
        values = self._values
        for i in range(0, len(state), 2):
            keys.append(state[i])
            values.append(state[i + 1])

    def _p_resolveConflict(self, s_old, s_com, s_new):
        b_old = type(self)()
        if s_old is not None:
            b_old.__setstate__(s_old)
        b_com = type(self)()
        if s_com is not None:
            b_com.__setstate__(s_com)
        b_new = type(self)()
        if s_new is not None:
            b_new.__setstate__(s_new)
        if (
            b_com._next != b_old._next or
            b_new._next != b_old._next
        ):
            raise BTreesConflictError(-1, -1, -1, 0)

        if not b_com or not b_new:
            raise BTreesConflictError(-1, -1, -1, 12)

        i_old = _SetIteration(b_old, True)
        i_com = _SetIteration(b_com, True)
        i_new = _SetIteration(b_new, True)

        def merge_error(reason):
            return BTreesConflictError(
                i_old.position, i_com.position, i_new.position, reason)

        result = type(self)()

        def merge_output(it):
            result._keys.append(it.key)
            result._values.append(it.value)
            it.advance()

        while i_old.active and i_com.active and i_new.active:
            cmpOC = compare(i_old.key, i_com.key)
            cmpON = compare(i_old.key, i_new.key)
            if cmpOC == 0:
                if cmpON == 0:
                    if i_com.value == i_old.value:
                        result[i_old.key] = i_new.value
                    elif i_new.value == i_old.value:
                        result[i_old.key] = i_com.value
                    else:
                        raise merge_error(1)
                    i_old.advance()
                    i_com.advance()
                    i_new.advance()
                elif (cmpON > 0):  # insert in new
                    merge_output(i_new)
                elif i_old.value == i_com.value:  # deleted new
                    if i_new.position == 1:
                        # Deleted the first item.  This will modify the
                        # parent node, so we don't know if merging will be
                        # safe
                        raise merge_error(13)
                    i_old.advance()
                    i_com.advance()
                else:
                    raise merge_error(2)
            elif cmpON == 0:
                if cmpOC > 0:  # insert committed
                    merge_output(i_com)
                elif i_old.value == i_new.value:  # delete committed
                    if i_com.position == 1:
                        # Deleted the first item.  This will modify the
                        # parent node, so we don't know if merging will be
                        # safe
                        raise merge_error(13)
                    i_old.advance()
                    i_new.advance()
                else:
                    raise merge_error(3)
            else:  # both keys changed
                cmpCN = compare(i_com.key, i_new.key)
                if cmpCN == 0:  # dueling insert
                    raise merge_error(4)
                if cmpOC > 0:  # insert committed
                    if cmpCN > 0:  # insert i_new first
                        merge_output(i_new)
                    else:
                        merge_output(i_com)
                elif cmpON > 0:  # insert i_new
                    merge_output(i_new)
                else:
                    raise merge_error(5)  # both deleted same key

        while i_com.active and i_new.active:  # new inserts
            cmpCN = compare(i_com.key, i_new.key)
            if cmpCN == 0:
                raise merge_error(6)  # dueling insert
            if cmpCN > 0:  # insert new
                merge_output(i_new)
            else:  # insert committed
                merge_output(i_com)

        while i_old.active and i_com.active:  # new deletes rest of original
            cmpOC = compare(i_old.key, i_com.key)
            if cmpOC > 0:  # insert committed
                merge_output(i_com)
            elif cmpOC == 0 and (i_old.value == i_com.value):  # del in new
                i_old.advance()
                i_com.advance()
            else:  # dueling deletes or delete and change
                raise merge_error(7)

        while i_old.active and i_new.active:
            # committed deletes rest of original
            cmpON = compare(i_old.key, i_new.key)
            if cmpON > 0:  # insert new
                merge_output(i_new)
            elif cmpON == 0 and (i_old.value == i_new.value):
                # deleted in committed
                i_old.advance()
                i_new.advance()
            else:  # dueling deletes or delete and change
                raise merge_error(8)

        if i_old.active:  # dueling deletes
            raise merge_error(9)

        while i_com.active:
            merge_output(i_com)

        while i_new.active:
            merge_output(i_new)

        if len(result._keys) == 0:  # pragma: no cover
            # If the output bucket is empty, conflict resolution doesn't have
            # enough info to unlink it from its containing BTree correctly.
            #
            # XXX TS, 2012-11-16:  I don't think this is possible
            #
            raise merge_error(10)

        result._next = b_old._next
        return result.__getstate__()

    def __repr__(self):
        return self._repr_helper(self.items())


class _MutableSetMixin:
    # Like _MutableMappingMixin, but for sets.
    def isdisjoint(self, other):
        """
        Return True if two sets have a null intersection.
        """
        for value in other:
            if value in self:
                return False
        return True

    def discard(self, key):
        """
        Remove an element from the set if it is a member.

        If not, do nothing and raise no exception.
        """
        # Written this way to avoid catching and accidentally
        # ignoring POSKeyError.
        if key in self:
            self.remove(key)

    def pop(self):
        """Return the popped value.  Raise KeyError if empty."""
        # Get our iter first to avoid catching and accidentally
        # ignoring POSKeyError
        it = iter(self)
        try:
            value = next(it)
        except StopIteration:
            raise KeyError
        self.discard(value)
        return value

    def __ior__(self, it):
        self.update(it)
        return self

    def __iand__(self, it):
        for value in (self - it):
            self.discard(value)
        return self

    def __isub__(self, it):
        if it is self:
            self.clear()
        else:
            for value in it:
                self.discard(value)
        return self

    def __ixor__(self, it):
        if it is self:
            self.clear()
        else:
            for value in it:
                if value in self:
                    self.discard(value)
                else:
                    self.add(value)
        return self


class Set(_MutableSetMixin, _BucketBase):

    __slots__ = ()

    def add(self, key):
        return self._set(self._to_key(key))[0]

    insert = add

    def remove(self, key):
        self._del(self._to_key(key))

    def update(self, items):
        add = self.add
        for i in items:
            add(i)

    def __getstate__(self):
        data = tuple(self._keys)
        if self._next is not None:
            return data, self._next
        return (data, )

    def __setstate__(self, state):
        if not isinstance(state[0], tuple):
            raise TypeError('tuple required for first state element')

        self.clear()
        if len(state) == 2:
            state, self._next = state
        else:
            self._next = None
            state = state[0]

        self._keys.extend(state)

    def _set(self, key, value=None, ifunset=False):
        index = self._search(key)
        if index < 0:
            index = -index - 1
            self._p_changed = True
            self._keys.insert(index, key)
            return True, None
        return False, None

    def _del(self, key):
        index = self._search(key)
        if index >= 0:
            self._p_changed = True
            del self._keys[index]
            return 0, 0
        raise KeyError(key)

    def __getitem__(self, i):
        return self._keys[i]

    def _split(self, index=-1):
        if index < 0 or index >= len(self._keys):
            index = len(self._keys) // 2
        new_instance = type(self)()
        new_instance._keys = self._keys[index:]
        del self._keys[index:]
        new_instance._next = self._next
        self._next = new_instance
        return new_instance

    def _p_resolveConflict(self, s_old, s_com, s_new):

        b_old = type(self)()
        if s_old is not None:
            b_old.__setstate__(s_old)
        b_com = type(self)()
        if s_com is not None:
            b_com.__setstate__(s_com)
        b_new = type(self)()
        if s_new is not None:
            b_new.__setstate__(s_new)

        if (
            b_com._next != b_old._next or
            b_new._next != b_old._next
        ):  # conflict: com or new changed _next
            raise BTreesConflictError(-1, -1, -1, 0)

        if not b_com or not b_new:  # conflict: com or new empty
            raise BTreesConflictError(-1, -1, -1, 12)

        i_old = _SetIteration(b_old, True)
        i_com = _SetIteration(b_com, True)
        i_new = _SetIteration(b_new, True)

        def merge_error(reason):
            return BTreesConflictError(
                i_old.position, i_com.position, i_new.position, reason)

        result = type(self)()

        def merge_output(it):
            result._keys.append(it.key)
            it.advance()

        while i_old.active and i_com.active and i_new.active:
            cmpOC = compare(i_old.key, i_com.key)
            cmpON = compare(i_old.key, i_new.key)
            if cmpOC == 0:
                if cmpON == 0:  # all match
                    merge_output(i_old)
                    i_com.advance()
                    i_new.advance()
                elif cmpON > 0:  # insert in new
                    merge_output(i_new)
                else:  # deleted new
                    if i_new.position == 1:
                        # Deleted the first item.  This will modify the
                        # parent node, so we don't know if merging will be
                        # safe
                        raise merge_error(13)
                    i_old.advance()
                    i_com.advance()
            elif cmpON == 0:
                if cmpOC > 0:  # insert committed
                    merge_output(i_com)
                else:  # delete committed
                    if i_com.position == 1:
                        # Deleted the first item.  This will modify the
                        # parent node, so we don't know if merging will be
                        # safe
                        raise merge_error(13)
                    i_old.advance()
                    i_new.advance()
            else:  # both com and new keys changed
                cmpCN = compare(i_com.key, i_new.key)
                if cmpCN == 0:  # both inserted same key
                    raise merge_error(4)
                if cmpOC > 0:  # insert committed
                    if cmpCN > 0:  # insert i_new first
                        merge_output(i_new)
                    else:
                        merge_output(i_com)
                elif cmpON > 0:  # insert i_new
                    merge_output(i_new)
                else:  # both com and new deleted same key
                    raise merge_error(5)

        while i_com.active and i_new.active:  # new inserts
            cmpCN = compare(i_com.key, i_new.key)
            if cmpCN == 0:  # dueling insert
                raise merge_error(6)
            if cmpCN > 0:  # insert new
                merge_output(i_new)
            else:  # insert committed
                merge_output(i_com)

        while i_old.active and i_com.active:  # new deletes rest of original
            cmpOC = compare(i_old.key, i_com.key)
            if cmpOC > 0:  # insert committed
                merge_output(i_com)
            elif cmpOC == 0:  # del in new
                i_old.advance()
                i_com.advance()
            else:  # dueling deletes or delete and change
                raise merge_error(7)

        while i_old.active and i_new.active:
            # committed deletes rest of original
            cmpON = compare(i_old.key, i_new.key)
            if cmpON > 0:  # insert new
                merge_output(i_new)
            elif cmpON == 0:  # deleted in committed
                i_old.advance()
                i_new.advance()
            else:  # dueling deletes or delete and change
                raise merge_error(8)

        if i_old.active:  # dueling deletes
            raise merge_error(9)

        while i_com.active:
            merge_output(i_com)

        while i_new.active:
            merge_output(i_new)

        if len(result._keys) == 0:  # pragma: no cover
            # If the output bucket is empty, conflict resolution doesn't have
            # enough info to unlink it from its containing BTree correctly.
            #
            # XXX TS, 2012-11-16:  I don't think this is possible
            #
            raise merge_error(10)

        result._next = b_old._next
        return result.__getstate__()

    def __repr__(self):
        return self._repr_helper(self._keys)


class _TreeItem:

    __slots__ = ('key', 'child')

    def __init__(self, key, child):
        self.key = key
        self.child = child


class _Tree(_ArithmeticMixin, _Base):

    __slots__ = ('_data', '_firstbucket')

    def __new__(cls, *args):
        value = _Base.__new__(cls, *args)
        # Empty trees don't get their __setstate__ called upon
        # unpickling (or __init__, obviously), so clear() is never called
        # and _data and _firstbucket are never defined, unless we do it here.
        value._data = []
        value._firstbucket = None
        return value

    def setdefault(self, key, value):
        return self._set(self._to_key(key), self._to_value(value), True)[1]

    def pop(self, key, default=_marker):
        try:
            return self._del(self._to_key(key))[1]
        except KeyError:
            if default is _marker:
                raise
            return default

    def update(self, items):
        if hasattr(items, 'iteritems'):
            items = items.iteritems()
        elif hasattr(items, 'items'):
            items = items.items()

        set = self.__setitem__
        for i in items:
            set(*i)

    def __setitem__(self, key, value):
        self._set(self._to_key(key), self._to_value(value))

    def __delitem__(self, key):
        self._del(self._to_key(key))

    def clear(self):
        if self._data:
            # In the case of __init__, this was already set by __new__
            self._data = []
        self._firstbucket = None

    def __bool__(self):
        return bool(self._data)

    def __len__(self):
        accumulated = 0
        bucket = self._firstbucket
        while bucket is not None:
            accumulated += len(bucket._keys)
            bucket = bucket._next
        return accumulated

    @property
    def size(self):
        return len(self._data)

    def _search(self, key):
        data = self._data
        if data:
            lo = 0
            hi = len(data)
            i = hi // 2
            while i > lo:
                cmp_ = compare(data[i].key, key)
                if cmp_ < 0:
                    lo = i
                elif cmp_ > 0:
                    hi = i
                else:
                    break
                i = (lo + hi) // 2
            return i
        return -1

    def _findbucket(self, key):
        index = self._search(key)
        if index >= 0:
            child = self._data[index].child
            if isinstance(child, self._bucket_type):
                return child
            return child._findbucket(key)

    def __contains__(self, key):
        try:
            tree_key = self._to_key(key)
        except TypeError:
            # Can't convert the key, so can't possibly be in the tree
            return False
        return key in (self._findbucket(tree_key) or ())

    def has_key(self, key):
        index = self._search(key)
        if index < 0:
            return False
        return self._data[index].child.has_key(key)

    def keys(self, min=_marker, max=_marker,
             excludemin=False, excludemax=False,
             itertype='iterkeys'):
        if not self._data:
            return ()

        if min is not _marker and min is not None:
            min = self._to_key(min)
            bucket = self._findbucket(min)
        else:
            bucket = self._firstbucket

        iterargs = min, max, excludemin, excludemax

        return _TreeItems(bucket, itertype, iterargs)

    def iterkeys(self, min=_marker, max=_marker,
                 excludemin=False, excludemax=False):
        return iter(self.keys(min, max, excludemin, excludemax))

    def __iter__(self):
        return iter(self.keys())

    def minKey(self, min=_marker):
        if min is _marker or min is None:
            bucket = self._firstbucket
        else:
            min = self._to_key(min)
            bucket = self._findbucket(min)
        if bucket is not None:
            return bucket.minKey(min)
        raise ValueError('empty tree')

    def maxKey(self, max=_marker):
        data = self._data
        if not data:
            raise ValueError('empty tree')
        if max is _marker or max is None:
            return data[-1].child.maxKey()

        max = self._to_key(max)
        index = self._search(max)
        if index and compare(data[index].child.minKey(), max) > 0:
            index -= 1  # pragma: no cover  no idea how to provoke this
        return data[index].child.maxKey(max)

    def _set(self, key, value=None, ifunset=False):
        if (
            self._p_jar is not None and
            self._p_oid is not None and
            self._p_serial is not None
        ):
            self._p_jar.readCurrent(self)
        data = self._data
        if data:
            index = self._search(key)
            child = data[index].child
        else:
            index = 0
            child = self._bucket_type()
            self._firstbucket = child
            data.append(_TreeItem(None, child))

        result = child._set(key, value, ifunset)
        grew = result[0]
        if grew:
            if type(child) is type(self):
                max_size = type(self).max_internal_size
            else:
                max_size = type(self).max_leaf_size
            if child.size > max_size:
                self._grow(child, index)

        # If a BTree contains only a single bucket, BTree.__getstate__()
        # includes the bucket's entire state, and the bucket doesn't get
        # an oid of its own.  So if we have a single oid-less bucket that
        # changed, it's *our* oid that should be marked as changed -- the
        # bucket doesn't have one.
        if (
            grew is not None and
            type(child) is self._bucket_type and
            len(data) == 1 and
            child._p_oid is None
        ):
            self._p_changed = 1
        return result

    def _grow(self, child, index):
        self._p_changed = True
        new_child = child._split()
        self._data.insert(index + 1, _TreeItem(new_child.minKey(), new_child))
        if len(self._data) >= type(self).max_internal_size * 2:
            self._split_root()

    def _split_root(self):
        child = type(self)()
        child._data = self._data
        child._firstbucket = self._firstbucket
        self._data = [_TreeItem(None, child)]
        self._grow(child, 0)

    def _split(self, index=None):
        data = self._data
        if index is None:
            index = len(data) // 2

        next = type(self)()
        next._data = data[index:]
        first = data[index]
        del data[index:]
        if len(data) == 0:
            self._firstbucket = None  # lost our bucket, can't buy no beer
        if isinstance(first.child, type(self)):
            next._firstbucket = first.child._firstbucket
        else:
            next._firstbucket = first.child
        return next

    def _del(self, key):
        if (
            self._p_jar is not None and
            self._p_oid is not None and
            self._p_serial is not None
        ):
            self._p_jar.readCurrent(self)

        data = self._data
        if not data:
            raise KeyError(key)

        index = self._search(key)
        child = data[index].child

        removed_first_bucket, value = child._del(key)

        # See comment in _set about small trees
        if (
            len(data) == 1 and
            type(child) is self._bucket_type and
            child._p_oid is None
        ):
            self._p_changed = True

        # fix up the node key, but not for the 0'th one.
        if index > 0 and child.size and compare(key, data[index].key) == 0:
            self._p_changed = True
            data[index].key = child.minKey()

        if removed_first_bucket:
            if index:
                data[index - 1].child._deleteNextBucket()
                removed_first_bucket = False  # clear flag
            else:
                self._firstbucket = child._firstbucket

        if not child.size:
            if type(child) is self._bucket_type:
                if index:
                    data[index - 1].child._deleteNextBucket()
                else:
                    self._firstbucket = child._next
                    removed_first_bucket = True
            del data[index]
            self._p_changed = True

        return removed_first_bucket, value

    def _deleteNextBucket(self):
        self._data[-1].child._deleteNextBucket()

    def __getstate__(self):
        data = self._data

        if not data:
            # Note: returning None here causes our __setstate__
            # to not be called on unpickling
            return None

        if (
            len(data) == 1 and
            type(data[0].child) is not type(self) and
            data[0].child._p_oid is None
        ):
            return ((data[0].child.__getstate__(), ), )

        data = iter(data)
        sdata = [next(data).child]
        for item in data:
            sdata.append(item.key)
            sdata.append(item.child)

        return tuple(sdata), self._firstbucket

    def __setstate__(self, state):
        if state and not isinstance(state[0], tuple):
            raise TypeError('tuple required for first state element')

        self.clear()
        if state is None:
            return

        if len(state) == 1:
            bucket = self._bucket_type()
            bucket.__setstate__(state[0][0])
            state = [bucket], bucket

        data, self._firstbucket = state
        data = list(reversed(data))

        # verify children are either tree or bucket nodes.
        # NOTE for tree-kind node type is compared as "is", not as
        # "isinstance", to match C version.
        for child in data[::2]:
            if not ((type(child) is type(self)) or
                    isinstance(child, self._bucket_type)):
                raise TypeError("tree child %s is neither %s nor %s" %
                                (_tp_name(type(child)), _tp_name(type(self)),
                                 _tp_name(self._bucket_type)))

        self._data.append(_TreeItem(None, data.pop()))
        while data:
            key = data.pop()
            child = data.pop()
            self._data.append(_TreeItem(key, child))

    def _assert(self, condition, message):
        if not condition:
            raise AssertionError(message)

    def _check(self, nextbucket=None):
        data = self._data
        assert_ = self._assert
        if not data:
            assert_(self._firstbucket is None,
                    "Empty BTree has non-NULL firstbucket")
            return
        assert_(self._firstbucket is not None,
                "Non-empty BTree has NULL firstbucket")

        child_class = type(data[0].child)
        for i in data:
            assert_(i.child is not None, "BTree has NULL child")
            assert_(type(i.child) is child_class,
                    "BTree children have different types")
            assert_(i.child.size, "Bucket length < 1")

        if child_class is type(self):
            assert_(
                self._firstbucket is data[0].child._firstbucket,
                "BTree has firstbucket different than "
                "its first child's firstbucket"
            )
            for i in range(len(data) - 1):
                data[i].child._check(data[i + 1].child._firstbucket)
            data[-1].child._check(nextbucket)
        elif child_class is self._bucket_type:
            assert_(
                self._firstbucket is data[0].child,
                "Bottom-level BTree node has inconsistent firstbucket belief"
            )
            for i in range(len(data) - 1):
                assert_(
                    data[i].child._next is data[i + 1].child,
                    "Bucket next pointer is damaged"
                )
            assert_(
                data[-1].child._next is nextbucket,
                "Bucket next pointer is damaged"
            )
        else:
            assert_(False, "Incorrect child type")

    def _p_resolveConflict(self, old, com, new):
        s_old = _get_simple_btree_bucket_state(old)
        s_com = _get_simple_btree_bucket_state(com)
        s_new = _get_simple_btree_bucket_state(new)
        return ((
            self._bucket_type()._p_resolveConflict(s_old, s_com, s_new), ), )

    def __repr__(self):
        r = super().__repr__()
        r = r.replace('Py', '')
        return r


def _get_simple_btree_bucket_state(state):
    if state is None:
        return state
    if not isinstance(state, tuple):
        raise TypeError("_p_resolveConflict: expected tuple or None for state")
    if len(state) == 2:  # non-degenerate BTree, can't resolve
        raise BTreesConflictError(-1, -1, -1, 11)
    # Peel away wrapper to get to only-bucket state.
    if len(state) != 1:
        raise TypeError("_p_resolveConflict: expected 1- or 2-tuple for state")
    state = state[0]
    if not isinstance(state, tuple) or len(state) != 1:
        raise TypeError("_p_resolveConflict: expected 1-tuple containing "
                        "bucket state")
    state = state[0]
    if not isinstance(state, tuple):
        raise TypeError("_p_resolveConflict: expected tuple for bucket state")
    return state


class _TreeItems:

    __slots__ = (
        'firstbucket',
        'itertype',
        'iterargs',
        'index',
        'it',
        'v',
        '_len',
    )

    def __init__(self, firstbucket, itertype, iterargs):
        self.firstbucket = firstbucket
        self.itertype = itertype
        self.iterargs = iterargs
        self.index = -1
        self.it = iter(self)
        self.v = None
        self._len = None

    def __getitem__(self, i):
        if isinstance(i, slice):
            return list(self)[i]
        if i < 0:
            i = len(self) + i
            if i < 0:
                raise IndexError(i)

        if i < self.index:
            self.index = -1
            self.it = iter(self)

        while i > self.index:
            try:
                self.v = next(self.it)
            except StopIteration:
                raise IndexError(i)
            else:
                self.index += 1
        return self.v

    def __len__(self):
        if self._len is None:
            i = 0
            for _ in self:
                i += 1
            self._len = i
        return self._len

    def __iter__(self):
        bucket = self.firstbucket
        itertype = self.itertype
        iterargs = self.iterargs
        done = 0
        # Note that we don't mind if the first bucket yields no
        # results due to an idiosyncrasy in how range searches are done.
        while bucket is not None:
            for k in getattr(bucket, itertype)(*iterargs):
                yield k
                done = 0
            if done:
                return
            bucket = bucket._next
            done = 1


class _TreeIterator:
    """ Faux implementation for BBB only.
    """

    def __init__(self, items):  # pragma: no cover
        raise TypeError(
            "TreeIterators are private implementation details "
            "of the C-based BTrees.\n\n"
            "Please use 'iter(tree)', rather than instantiating "
            "one directly."
        )


class Tree(_MutableMappingMixin, _Tree):

    __slots__ = ()

    def get(self, key, default=None):
        bucket = self._findbucket(key)
        if bucket:
            return bucket.get(key, default)
        return default

    def __getitem__(self, key):
        bucket = self._findbucket(key)
        if bucket:
            return bucket[key]
        raise KeyError(key)

    def values(self, min=_marker, max=_marker,
               excludemin=False, excludemax=False):
        return self.keys(min, max, excludemin, excludemax, 'itervalues')

    def itervalues(self, min=_marker, max=_marker,
                   excludemin=False, excludemax=False):
        return iter(self.values(min, max, excludemin, excludemax))

    def items(self, min=_marker, max=_marker,
              excludemin=False, excludemax=False):
        return self.keys(min, max, excludemin, excludemax, 'iteritems')

    def iteritems(self, min=_marker, max=_marker,
                  excludemin=False, excludemax=False):
        return iter(self.items(min, max, excludemin, excludemax))

    def byValue(self, min):
        return reversed(
            sorted((v, k) for (k, v) in self.iteritems() if v >= min))

    def insert(self, key, value):
        return bool(self._set(key, value, True)[0])


class TreeSet(_MutableSetMixin, _Tree):

    __slots__ = ()

    def add(self, key):
        return self._set(self._to_key(key))[0]

    insert = add

    def remove(self, key):
        self._del(self._to_key(key))

    def update(self, items):
        add = self.add
        for i in items:
            add(i)

    _p_resolveConflict = _Tree._p_resolveConflict


class set_operation:

    __slots__ = (
        'func',
        'set_type',
        '__name__',
        '_module',
    )

    def __init__(self, func, set_type):
        self.func = func
        self.set_type = set_type
        self.__name__ = func.__name__
        self._module = func.__module__

    __module__ = property(
        lambda self: self._module,
        lambda self, nv: setattr(self, '_module', nv)
    )

    def __call__(self, *a, **k):
        return self.func(self.set_type, *a, **k)


def difference(set_type, o1, o2):
    if o1 is None or o2 is None:
        return o1
    i1 = _SetIteration(o1, True, 0)
    i2 = _SetIteration(o2, False, 0, True)
    if i1.useValues:
        result = o1._mapping_type()

        def copy(i):
            result._keys.append(i.key)
            result._values.append(i.value)

    else:
        result = o1._set_type()

        def copy(i):
            result._keys.append(i.key)

    while i1.active and i2.active:
        cmp_ = compare(i1.key, i2.key)
        if cmp_ < 0:
            copy(i1)
            i1.advance()
        elif cmp_ == 0:
            i1.advance()
            i2.advance()
        else:
            i2.advance()
    while i1.active:
        copy(i1)
        i1.advance()
    return result


def union(set_type, o1, o2):
    if o1 is None:
        return o2
    if o2 is None:
        return o1
    i1 = _SetIteration(o1, False, 0, True)
    i2 = _SetIteration(o2, False, 0, True)
    result = set_type()

    def copy(i):
        result._keys.append(i.key)

    while i1.active and i2.active:
        cmp_ = compare(i1.key, i2.key)
        if cmp_ < 0:
            copy(i1)
            i1.advance()
        elif cmp_ == 0:
            copy(i1)
            i1.advance()
            i2.advance()
        else:
            copy(i2)
            i2.advance()
    while i1.active:
        copy(i1)
        i1.advance()
    while i2.active:
        copy(i2)
        i2.advance()
    return result


def intersection(set_type, o1, o2):
    if o1 is None:
        return o2
    if o2 is None:
        return o1
    i1 = _SetIteration(o1, False, 0, True)
    i2 = _SetIteration(o2, False, 0, True)
    result = set_type()

    def copy(i):
        result._keys.append(i.key)

    while i1.active and i2.active:
        cmp_ = compare(i1.key, i2.key)
        if cmp_ < 0:
            i1.advance()
        elif cmp_ == 0:
            copy(i1)
            i1.advance()
            i2.advance()
        else:
            i2.advance()
    return result


def _prepMergeIterators(o1, o2):
    MERGE_DEFAULT = getattr(o1, 'MERGE_DEFAULT', None)
    if MERGE_DEFAULT is None:
        raise TypeError("invalid set operation")
    i1 = _SetIteration(o1, True, MERGE_DEFAULT)
    i2 = _SetIteration(o2, True, MERGE_DEFAULT)
    return i1, i2


def weightedUnion(set_type, o1, o2, w1=1, w2=1):
    if o1 is None:
        if o2 is None:
            return 0, None
        return w2, o2
    if o2 is None:
        return w1, o1
    i1, i2 = _prepMergeIterators(o1, o2)
    MERGE = getattr(o1, 'MERGE', None)
    if MERGE is None and i1.useValues and i2.useValues:
        raise TypeError("invalid set operation")
    MERGE_WEIGHT = getattr(o1, 'MERGE_WEIGHT')
    if (not i1.useValues) and i2.useValues:
        i1, i2 = i2, i1
        w1, w2 = w2, w1
    _merging = i1.useValues or i2.useValues

    if _merging:
        result = o1._mapping_type()

        def copy(i, w):
            result._keys.append(i.key)
            result._values.append(MERGE_WEIGHT(i.value, w))

    else:
        result = o1._set_type()

        def copy(i, w):
            result._keys.append(i.key)

    while i1.active and i2.active:
        cmp_ = compare(i1.key, i2.key)
        if cmp_ < 0:
            copy(i1, w1)
            i1.advance()
        elif cmp_ == 0:
            result._keys.append(i1.key)
            if _merging:
                result._values.append(MERGE(i1.value, w1, i2.value, w2))
            i1.advance()
            i2.advance()
        else:
            copy(i2, w2)
            i2.advance()
    while i1.active:
        copy(i1, w1)
        i1.advance()
    while i2.active:
        copy(i2, w2)
        i2.advance()
    return 1, result


def weightedIntersection(set_type, o1, o2, w1=1, w2=1):
    if o1 is None:
        if o2 is None:
            return 0, None
        return w2, o2
    if o2 is None:
        return w1, o1
    i1, i2 = _prepMergeIterators(o1, o2)
    MERGE = getattr(o1, 'MERGE', None)
    if MERGE is None and i1.useValues and i2.useValues:
        raise TypeError("invalid set operation")
    if (not i1.useValues) and i2.useValues:
        i1, i2 = i2, i1
        w1, w2 = w2, w1
    _merging = i1.useValues or i2.useValues
    if _merging:
        result = o1._mapping_type()
    else:
        result = o1._set_type()
    while i1.active and i2.active:
        cmp_ = compare(i1.key, i2.key)
        if cmp_ < 0:
            i1.advance()
        elif cmp_ == 0:
            result._keys.append(i1.key)
            if _merging:
                result._values.append(MERGE(i1.value, w1, i2.value, w2))
            i1.advance()
            i2.advance()
        else:
            i2.advance()
    if isinstance(result, (Set, TreeSet)):
        return w1 + w2, result
    return 1, result


def multiunion(set_type, seqs):
    # XXX simple/slow implementation. Goal is just to get tests to pass.
    result = set_type()
    for s in seqs:
        try:
            iter(s)
        except TypeError:
            s = set_type((s, ))
        result.update(s)
    return result


def MERGE(self, value1, weight1, value2, weight2):
    return (value1 * weight1) + (value2 * weight2)


def MERGE_WEIGHT_default(self, value, weight):
    return value


def MERGE_WEIGHT_numeric(self, value, weight):
    return value * weight


def _fix_pickle(mod_dict, mod_name):
    # Make the pure-Python objects pickle with the same
    # class names and types as the C extensions by setting the appropriate
    # _BTree_reduce_as attribute.
    # If the C extensions are not available, we also change the
    # __name__ attribute of the type to match the C name (otherwise
    # we wind up with *Py in the pickles)
    # Each module must call this as `_fix_pickle(globals(), __name__)`
    # at the bottom.

    mod_prefix = mod_name.split('.')[-1][:2]  # BTrees.OOBTree -> 'OO'
    bucket_name = mod_prefix + 'Bucket'
    py_bucket_name = bucket_name + 'Py'

    have_c_extensions = mod_dict[bucket_name] is not mod_dict[py_bucket_name]

    for name in 'Bucket', 'Set', 'BTree', 'TreeSet', 'TreeIterator':
        raw_name = mod_prefix + name
        py_name = raw_name + 'Py'
        try:
            py_type = mod_dict[py_name]
        except KeyError:
            if name == 'TreeIterator':
                # Optional
                continue
            raise  # pragma: no cover
        raw_type = mod_dict[raw_name]  # Could be C or Python

        py_type._BTree_reduce_as = raw_type
        py_type._BTree_reduce_up_bound = py_type

        if not have_c_extensions:  # pragma: no cover
            # Set FooPy to have the __name__ of simply Foo.
            # We can't do this if the C extension is available,
            # because then mod_dict[FooPy.__name__] is not FooPy
            # and pickle refuses to save something like that.
            # On the other hand (no C extension) this makes our
            # Python pickle match the C version by default
            py_type.__name__ = raw_name
            py_type.__qualname__ = raw_name  # Py 3.3+


# tp_name returns full name of a type in the same way as how it is provided by
# typ->tp_name in C.
def _tp_name(typ):
    return '.'.join([typ.__module__, typ.__name__])



##############################################################################
#
# Copyright Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Support functions to eliminate the boilerplate involved in defining
BTree modules.
"""
import sys

from zope.interface import classImplements
from zope.interface import directlyProvides


def _create_classes(
        module_name, key_datatype, value_datatype,
):
    from ._base import MERGE  # Won't always want this.
    from ._base import Bucket
    from ._base import Set
    from ._base import Tree
    from ._base import TreeSet
    from ._base import _TreeItems as TreeItems
    from ._base import _TreeIterator

    classes = {}

    prefix = key_datatype.prefix_code + value_datatype.prefix_code

    classes['TreeItems'] = classes['TreeItemsPy'] = TreeItems
    for base in (
            Bucket,
            Set,
            (Tree, 'BTree'),
            TreeSet,
            (_TreeIterator, 'TreeIterator'),
    ):
        if isinstance(base, tuple):
            base, base_name = base
        else:
            base_name = base.__name__

        # XXX: Consider defining these with their natural names
        # now and only aliasing them to 'Py' instead of the
        # opposite. That should make pickling easier.
        name = prefix + base_name + 'Py'
        cls = type(name, (base,), dict(
            _to_key=key_datatype,
            _to_value=value_datatype,
            MERGE=MERGE,
            MERGE_WEIGHT=value_datatype.apply_weight,
            MERGE_DEFAULT=value_datatype.multiplication_identity,
            # max_leaf_size and max_internal_size are set
            # for BTree and TreeSet later, when we do the same thing
            # for C.
        ))
        cls.__module__ = module_name
        key_datatype.add_extra_methods(base_name, cls)

        classes[cls.__name__] = cls
        # Importing the C extension does this for the non-py
        # classes.
        # TODO: Unify that.
        classes[base_name + 'Py'] = cls

    for cls in classes.values():
        cls._mapping_type = classes['BucketPy']
        cls._set_type = classes['SetPy']

        if 'Set' in cls.__name__:
            cls._bucket_type = classes['SetPy']
        else:
            cls._bucket_type = classes['BucketPy']

    return classes


def _create_set_operations(module_name, key_type, value_type, set_type):
    from ._base import difference
    from ._base import intersection
    from ._base import multiunion
    from ._base import set_operation
    from ._base import union
    from ._base import weightedIntersection
    from ._base import weightedUnion

    ops = {
        op.__name__ + 'Py': set_operation(op, set_type)
        for op in (
            difference, intersection,
            union,
        ) + (
            (weightedIntersection, weightedUnion,)
            if value_type.supports_value_union()
            else ()
        ) + (
            (multiunion,)
            if key_type.supports_value_union()
            else ()
        )
    }

    for key, op in ops.items():
        op.__module__ = module_name
        op.__name__ = key

    # TODO: Pickling. These things should be looked up by name.
    return ops


def _create_globals(module_name, key_datatype, value_datatype):
    classes = _create_classes(module_name, key_datatype, value_datatype)
    set_type = classes['SetPy']
    set_ops = _create_set_operations(
        module_name, key_datatype, value_datatype, set_type,
    )

    classes.update(set_ops)
    return classes


def populate_module(mod_globals,
                    key_datatype, value_datatype,
                    interface, module=None):
    import collections.abc

    from . import Interfaces as interfaces
    from ._base import _fix_pickle
    from ._compat import import_c_extension

    module_name = mod_globals['__name__']
    # Define the Python implementations
    mod_globals.update(
        _create_globals(module_name, key_datatype, value_datatype)
    )
    # Import the C versions, if possible. Whether or not this is possible,
    # this currently makes the non-`Py' suffixed names available. This should
    # change if we start defining the Python classes with their natural name,
    # only aliased to the 'Py` suffix (which simplifies pickling)
    import_c_extension(mod_globals)

    # Next, define __all__ after all the name aliasing is done.
    # XXX: Maybe derive this from the values we create.
    mod_all = (
        'Bucket', 'Set', 'BTree', 'TreeSet',
        'union', 'intersection', 'difference',
        'weightedUnion', 'weightedIntersection', 'multiunion',
    )
    prefix = key_datatype.prefix_code + value_datatype.prefix_code

    mod_all += tuple(prefix + c for c in ('Bucket', 'Set', 'BTree', 'TreeSet'))

    mod_globals['__all__'] = tuple(c for c in mod_all if c in mod_globals)

    mod_globals['using64bits'] = (
        key_datatype.using64bits or value_datatype.using64bits
    )

    # XXX: We can probably do better than fix_pickle now;
    # we can know if we're going to be renaming classes
    # ahead of time. See above.
    _fix_pickle(mod_globals, module_name)

    # Apply interface definitions.
    directlyProvides(module or sys.modules[module_name], interface)
    for cls_name, iface in {
            'BTree': interfaces.IBTree,
            'Bucket': interfaces.IMinimalDictionary,
            'Set': interfaces.ISet,
            'TreeSet': interfaces.ITreeSet,
            'TreeItems': interfaces.IMinimalSequence,
    }.items():
        classImplements(mod_globals[cls_name], iface)
        classImplements(mod_globals[cls_name + 'Py'], iface)

    for cls_name, abc in {
            'BTree': collections.abc.MutableMapping,
            'Bucket': collections.abc.MutableMapping,
            'Set': collections.abc.MutableSet,
            'TreeSet': collections.abc.MutableSet,
    }.items():
        abc.register(mod_globals[cls_name])
        # Because of some quirks in the implementation of
        # ABCMeta.__instancecheck__, and the shenanigans we currently do to
        # make Python classes pickle without the 'Py' suffix, it's not actually
        # necessary to register the Python version of the class. Specifically,
        # ABCMeta asks for the object's ``__class__`` instead of using
        # ``type()``, and our objects have a ``@property`` for ``__class__``
        # that returns the C version.
        #
        # That's too many coincidences to rely on though.
        abc.register(mod_globals[cls_name + 'Py'])

    # Set node sizes.
    for cls_name in ('BTree', 'TreeSet'):

        for suffix in ('', 'Py'):
            cls = mod_globals[cls_name + suffix]
            cls.max_leaf_size = key_datatype.bucket_size_for_value(
                value_datatype
            )
            cls.max_internal_size = key_datatype.tree_size


def create_module(prefix):
    import types

    from . import Interfaces
    from . import _datatypes as datatypes

    mod = types.ModuleType('BTrees.' + prefix + 'BTree')

    key_type = getattr(datatypes, prefix[0])()
    val_type = getattr(datatypes, prefix[1])().as_value_type()

    iface_name = 'I' + key_type.long_name + val_type.long_name + 'BTreeModule'

    iface = getattr(Interfaces, iface_name)

    populate_module(vars(mod), key_type, val_type, iface, mod)
    return mod



##############################################################################
#
# Copyright Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Descriptions of the datatypes supported by this package.
"""

import abc
import operator
import struct

from .utils import Lazy


class DataType:
    """
    Describes a data type used as a value.

    Subclasses will be defined for each particular
    supported type.
    """

    # The name for this datatype as used in interface names.
    long_name = None

    # The prefix code for this data type. Usually a single letter.
    prefix_code = None

    # The multiplication identity for this data type. Used in
    # combining (merging) data types. Leave undefined if this is
    # not a valid operation.
    multiplication_identity = None

    # Does the data take up 64-bits? Currently only relevant for the
    # integer key types.
    using64bits = False

    def __init__(self):
        if not self.prefix_code:
            self.prefix_code = type(self).__name__

    def __call__(self, item):
        """
        Verify *item* is in the correct format (or "close" enough)
        and return the item or its suitable conversion.

        If this cannot be done, raise a :exc:`TypeError`.

        The definition of "close" varies according to the datatypes.
        For example, integer datatypes will accept anything that can
        be converted into an integer using normal python coercion
        rules (calling ``__index__``) and where the integer fits into
        the required native type size (e.g., 4 bytes).
        """
        raise NotImplementedError

    def coerce(self, item):
        """
        Coerce *item* into something that can be used with
        ``__call__`` and return it.

        The coercion rules will vary by datatype. This exists only
        for test cases. The default is to perform the same validation
        as ``__call__``.
        """
        return self(item)

    def apply_weight(self, item, weight):
        """
        Apply a *weight* multiplier to *item*.

        Used when merging data structures. The *item* will be a
        value.
        """
        return item

    def as_value_type(self):
        # Because ``O'`` is used for both key and value,
        # we can override this to get the less restrictive value type.
        return self

    def supports_value_union(self):
        raise NotImplementedError

    def getTwoExamples(self):
        """
        Provide two distinct (non equal) examples acceptable to `__call__`.

        This is for testing.
        """
        return "object1", "object2"

    def get_lower_bound(self):
        """
        If there is a lower bound (inclusive) on the data type, return
        it. Otherwise, return ``None``.

        For integer types, this will only depend on whether it
        supports signed or unsigned values, and the answer will be 0
        or a negative number. For object types, ``None`` is always
        defined to sort as the lowest bound.

        This can be relevant for both key and value types.
        """
        return None

    def get_upper_bound(self):
        """
        If there is an upper bound (inclusive) on the data type,
        return it. Otherwise, return ``None``.

        Remarks are as for `get_lower_bound`.
        """
        return None

    def add_extra_methods(self, base_name, cls):
        """
        Hook method called on the key datatype to add zero or more
        desired arbitrary additional, non-standard, methods to the
        *cls* being constructed.

        *base_name* will be a string identifying the particular family
        of class being constructed, such as 'Bucket' or 'BTree'.
        """


class KeyDataType(DataType):
    """
    Describes a data type that has additional restrictions allowing it
    to be used as a key.
    """

    # When used as the key, this number determines the
    # max_internal_size.
    tree_size = 500

    default_bucket_size = 120

    def __call__(self, item):
        raise NotImplementedError

    def bucket_size_for_value(self, value_type):
        """
        What should the bucket (``max_leaf_size``) be when
        this data type is used with the given *value_type*?
        """
        if isinstance(value_type, Any):
            return self.default_bucket_size // 2
        return self.default_bucket_size


class Any(DataType):
    """
    Arbitrary Python objects.
    """
    prefix_code = 'O'
    long_name = 'Object'

    def __call__(self, item):
        return item

    def supports_value_union(self):
        return False


class _HasDefaultComparison(abc.ABC):
    """
    An `abc.ABC <https://docs.python.org/3/library/abc.html>_` for
    checking whether an item has default comparison.

    All we have to do is override ``__subclasshook__`` to implement an
    algorithm determining whether a class has default comparison.
    Python and the ABC machinery will take care of translating
    ``isinstance(thing, _HasDefaultComparison)`` into something like
    ``_HasDefaultComparison.__subclasshook__(type(thing))``. The ABC
    handles caching the answer (based on exact classes, no MRO), and
    getting the type from ``thing`` (including mostly dealing with
    old-style) classes on Python 2.
    """

    # Comparisons only use special methods defined on the
    # type, not instance variables.

    # On CPython 3, classes inherit __lt__ with ``__objclass__`` of ``object``.
    # On PyPy3, they do.
    #
    # Test these conditions at runtime and define the method variant
    # appropriately.
    #
    # Remember the method is checking if the object has default comparison
    assert '__lt__' not in DataType.__dict__
    if getattr(DataType.__lt__, '__objclass__', None) is object:
        # CPython 3
        @classmethod
        def __subclasshook__(cls, C, _NoneType=type(None)):
            if C is _NoneType:
                return False
            defining_class = getattr(C.__lt__, '__objclass__', None)
            if defining_class is None:
                # Implemented in Python
                return False
            return C.__lt__.__objclass__ is object
    else:
        # PyPy3
        @classmethod
        def __subclasshook__(
            cls, C, _object_lt=object.__lt__, _NoneType=type(None)
        ):
            if C is _NoneType:
                return False
            return C.__lt__ is _object_lt


class O(KeyDataType):  # noqa E742
    """
    Arbitrary (sortable) Python objects.
    """
    long_name = 'Object'
    tree_size = 250
    default_bucket_size = 60

    def as_value_type(self):
        return Any()

    def supports_value_union(self):
        return False

    def __call__(self, item):
        if isinstance(item, _HasDefaultComparison):
            raise TypeError(
                "Object of class {} has default comparison".format(
                    type(item).__name__
                )
            )
        return item


class _AbstractNativeDataType(KeyDataType):
    """
    Uses `struct.Struct` to verify that the data can fit into a native
    type.
    """

    _struct_format = None
    _as_python_type = NotImplementedError
    _required_python_type = object
    _error_description = None
    _as_packable = operator.index

    @Lazy
    def _check_native(self):
        return struct.Struct(self._struct_format).pack

    def __call__(self, item):
        try:
            self._check_native(self._as_packable(item))
        except (struct.error, TypeError, ValueError):
            # PyPy can raise ValueError converting a negative number to a
            # unsigned value.
            if isinstance(item, int):
                raise TypeError("Value out of range", item)
            raise TypeError(self._error_description)

        return self._as_python_type(item)

    def apply_weight(self, item, weight):
        return item * weight

    def supports_value_union(self):
        return True


class _AbstractIntDataType(_AbstractNativeDataType):
    _as_python_type = int
    _required_python_type = int
    multiplication_identity = 1
    long_name = "Integer"

    def getTwoExamples(self):
        return 1, 2

    def get_lower_bound(self):
        exp = 64 if self.using64bits else 32
        exp -= 1
        return int(-(2 ** exp))

    def get_upper_bound(self):
        exp = 64 if self.using64bits else 32
        exp -= 1
        return int(2 ** exp - 1)


class _AbstractUIntDataType(_AbstractIntDataType):
    long_name = 'Unsigned'

    def get_lower_bound(self):
        return 0

    def get_upper_bound(self):
        exp = 64 if self.using64bits else 32
        return int(2 ** exp - 1)


class I(_AbstractIntDataType):  # noqa E742
    _struct_format = 'i'
    _error_description = "32-bit integer expected"


class U(_AbstractUIntDataType):
    _struct_format = 'I'
    _error_description = 'non-negative 32-bit integer expected'


class F(_AbstractNativeDataType):
    _struct_format = 'f'
    _as_python_type = float
    _error_description = 'float expected'
    multiplication_identity = 1.0
    long_name = 'Float'

    def _as_packable(self, k):  # identity
        return k

    def getTwoExamples(self):
        return 0.5, 1.5


class L(_AbstractIntDataType):
    _struct_format = 'q'
    _error_description = '64-bit integer expected'
    using64bits = True


class Q(_AbstractUIntDataType):
    _struct_format = 'Q'
    _error_description = 'non-negative 64-bit integer expected'
    using64bits = True


class _AbstractBytes(KeyDataType):
    """
    An exact-length byte string type.

    This must be subclassed to provide the actual byte length.
    """
    tree_size = 500
    default_bucket_size = 500
    _length = None

    def __call__(self, item):
        if not isinstance(item, bytes) or len(item) != self._length:
            raise TypeError(
                f"{self._length}-byte array expected, not {item!r}"
            )
        return item

    def supports_value_union(self):
        # We don't implement 'multiunion' for fsBTree.
        return False


class f(_AbstractBytes):
    """
    The key type for an ``fs`` tree.

    This is a two-byte prefix of an overall 8-byte value
    like a ZODB object ID or transaction ID.
    """

    # Our keys are treated like integers; the module
    # implements IIntegerObjectBTreeModule
    long_name = 'Integer'
    prefix_code = 'f'
    _length = 2

    # Check it can be converted to a two-byte
    # value. Note that even though we allow negative values
    # that can break test assumptions: -1 < 0 < 1, but the byte
    # values for those are \xff\xff > \x00\x00 < \x00\x01.
    _as_2_bytes = struct.Struct('>h').pack

    def coerce(self, item):
        try:
            return self(item)
        except TypeError:
            try:
                return self._as_2_bytes(operator.index(item))
            except struct.error as e:
                raise TypeError(e)

    @staticmethod
    def _make_Bucket_toString():
        def toString(self):
            return b''.join(self._keys) + b''.join(self._values)
        return toString

    @staticmethod
    def _make_Bucket_fromString():
        def fromString(self, v):
            length = len(v)
            if length % 8 != 0:
                raise ValueError()
            count = length // 8
            keys, values = v[:count * 2], v[count * 2:]
            self.clear()
            while keys and values:
                key, keys = keys[:2], keys[2:]
                value, values = values[:6], values[6:]
                self._keys.append(key)
                self._values.append(value)
            return self
        return fromString

    def add_extra_methods(self, base_name, cls):
        if base_name == 'Bucket':
            cls.toString = self._make_Bucket_toString()
            cls.fromString = self._make_Bucket_fromString()


class s(_AbstractBytes):
    """
    The value type for an ``fs`` tree.

    This is a 6-byte suffix of an overall 8-byte value
    like a ZODB object ID or transaction ID.
    """

    # Our values are treated like objects; the
    # module implements IIntegerObjectBTreeModule
    long_name = 'Object'
    prefix_code = 's'
    _length = 6

    def get_lower_bound(self):
        # Negative values have the high bit set, which is incompatible
        # with our transformation.
        return 0

    # To coerce an integer, as used in tests, first convert to 8 bytes
    # in big-endian order, then ensure the first two
    # are 0 and cut them off.
    _as_8_bytes = struct.Struct('>q').pack

    def coerce(self, item):
        try:
            return self(item)
        except TypeError:
            try:
                as_bytes = self._as_8_bytes(operator.index(item))
            except struct.error as e:
                raise TypeError(e)

            if as_bytes[:2] != b'\x00\x00':
                raise TypeError(
                    "Cannot convert {!r} to 6 bytes ({!r})".format(
                        item, as_bytes
                    )
                )
            return as_bytes[2:]



##############################################################################
#
# Copyright (c) 2001-2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
from BTrees import OOBTree

from ._test_builder import update_module
from .common import BTreeTests


class OOBTreeTest(BTreeTests):

    def test_byValue(self):
        ITEMS = [(y, x) for x, y in enumerate('abcdefghijklmnopqrstuvwxyz')]
        tree = self._makeOne(ITEMS)
        self.assertEqual(list(tree.byValue(22)),
                         [(y, x) for x, y in reversed(ITEMS[22:])])

    def testRejectDefaultComparisonOnSet(self):
        # Check that passing in keys w default comparison fails. Only
        # applies to new-style class instances if we're using the C
        # extensions; old-style instances are too hard to introspect
        # in C.

        # This is white box because we know that the check is being
        # used in a function that's used in lots of places.
        # Otherwise, there are many permutations that would have to be
        # checked.
        t = self._makeOne()

        class OldStyle:
            pass

        if self._getTargetClass() is OOBTree.OOBTreePy:
            with self.assertRaises(TypeError):
                t[OldStyle()] = 1

        class C:
            pass

        with self.assertRaises(TypeError) as raising:
            t[C()] = 1

        self.assertEqual(
            raising.exception.args[0],
            "Object of class C has default comparison")

        class With___lt__:
            def __lt__(*args):
                return 1

        c = With___lt__()
        t[c] = 1
        t.clear()

        class With___lt__Old:
            def __lt__(*args):
                return 1

        c = With___lt__Old()
        t[c] = 1

        t.clear()

    def testAcceptDefaultComparisonOnGet(self):
        # Issue #42
        t = self._makeOne()

        class C:
            pass

        self.assertEqual(t.get(C(), 42), 42)
        self.assertRaises(KeyError, t.__getitem__, C())
        self.assertNotIn(C(), t)

    def testNewStyleClassWithCustomMetaClassAllowed(self):
        class Meta(type):
            def __lt__(cls, other):
                return 1

        cls = Meta('Class', (object,), {})
        m = self._makeOne()
        m[cls] = self.getTwoValues()[0]

    def test_None_is_smallest(self):
        t = self._makeOne()
        for i in range(999):  # Make sure we multiple buckets
            t[i] = i * i
        t[None] = -1
        for i in range(-99, 0):  # Make sure we multiple buckets
            t[i] = i * i
        self.assertEqual(list(t), [None] + list(range(-99, 999)))
        self.assertEqual(list(t.values()),
                         [-1] + [i * i for i in range(-99, 999)])
        self.assertEqual(t[2], 4)
        self.assertEqual(t[-2], 4)
        self.assertEqual(t[None], -1)
        t[None] = -2
        self.assertEqual(t[None], -2)
        t2 = t.__class__(t)
        del t[None]
        self.assertEqual(list(t), list(range(-99, 999)))

        if 'Py' in self.__class__.__name__:
            return
        from BTrees.OOBTree import difference
        from BTrees.OOBTree import intersection
        from BTrees.OOBTree import union
        self.assertEqual(list(difference(t2, t).items()), [(None, -2)])
        self.assertEqual(list(union(t, t2)), list(t2))
        self.assertEqual(list(intersection(t, t2)), list(t))

    def testDeleteNoneKey(self):
        # Check that a None key can be deleted in Python 2.
        # This doesn't work on Python 3 because None is unorderable,
        # so the tree can't be searched. But None also can't be inserted,
        # and we don't support migrating Python 2 databases to Python 3.
        t = self._makeOne()
        bucket_state = ((None, 42),)
        tree_state = ((bucket_state,),)
        t.__setstate__(tree_state)

        self.assertEqual(t[None], 42)
        del t[None]

    def testUnpickleNoneKey(self):
        # All versions (py2 and py3, C and Python) can unpickle
        # data that looks like this: {None: 42}, even though None
        # is unorderable..
        # This pickle was captured in BTree/ZODB3 3.10.7
        import pickle

        data = (
            b'ccopy_reg\n__newobj__\np0\n('
            b'cBTrees.OOBTree\nOOBTree\np1\ntp2\nRp3\n('
            b'(((NI42\ntp4\ntp5\ntp6\ntp7\nb.'
        )

        t = pickle.loads(data)
        keys = list(t)
        self.assertEqual([None], keys)

    def testIdentityTrumpsBrokenComparison(self):
        # Identical keys always match, even if their comparison is
        # broken. See https://github.com/zopefoundation/BTrees/issues/50
        from functools import total_ordering

        @total_ordering
        class Bad:
            def __eq__(self, other):
                return False

            __lt__ = __cmp__ = __eq__

        t = self._makeOne()
        bad_key = Bad()
        t[bad_key] = 42

        self.assertIn(bad_key, t)
        self.assertEqual(list(t), [bad_key])

        del t[bad_key]
        self.assertNotIn(bad_key, t)
        self.assertEqual(list(t), [])


update_module(globals(), OOBTree, btree_tests_base=OOBTreeTest)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import unittest

from BTrees.tests.common import permutations


class DegenerateBTree(unittest.TestCase):
    # Build a degenerate tree (set).  Boxes are BTree nodes.  There are
    # 5 leaf buckets, each containing a single int.  Keys in the BTree
    # nodes don't appear in the buckets.  Seven BTree nodes are purely
    # indirection nodes (no keys).  Buckets aren't all at the same depth:
    #
    #     +------------------------+
    #     |          4             |
    #     +------------------------+
    #         |              |
    #         |              v
    #         |             +-+
    #         |             | |
    #         |             +-+
    #         |              |
    #         v              v
    #     +-------+   +-------------+
    #     |   2   |   |   6     10  |
    #     +-------+   +-------------+
    #      |     |     |     |     |
    #      v     v     v     v     v
    #     +-+   +-+   +-+   +-+   +-+
    #     | |   | |   | |   | |   | |
    #     +-+   +-+   +-+   +-+   +-+
    #      |     |     |     |     |
    #      v     v     v     v     v
    #      1     3    +-+    7     11
    #                 | |
    #                 +-+
    #                  |
    #                  v
    #                  5
    #
    # This is nasty for many algorithms.  Consider a high-end range search
    # for 4.  The BTree nodes direct it to the 5 bucket, but the correct
    # answer is the 3 bucket, which requires going in a different direction
    # at the very top node already.  Consider a low-end range search for
    # 9.  The BTree nodes direct it to the 7 bucket, but the correct answer
    # is the 11 bucket.  This is also a nasty-case tree for deletions.

    def _build_degenerate_tree(self):
        # Build the buckets and chain them together.
        from BTrees.check import check
        from BTrees.IIBTree import IISet
        from BTrees.IIBTree import IITreeSet
        bucket11 = IISet([11])

        bucket7 = IISet()
        bucket7.__setstate__(((7,), bucket11))

        bucket5 = IISet()
        bucket5.__setstate__(((5,), bucket7))

        bucket3 = IISet()
        bucket3.__setstate__(((3,), bucket5))

        bucket1 = IISet()
        bucket1.__setstate__(((1,), bucket3))

        # Build the deepest layers of indirection nodes.
        ts = IITreeSet
        tree1 = ts()
        tree1.__setstate__(((bucket1,), bucket1))

        tree3 = ts()
        tree3.__setstate__(((bucket3,), bucket3))

        tree5lower = ts()
        tree5lower.__setstate__(((bucket5,), bucket5))
        tree5 = ts()
        tree5.__setstate__(((tree5lower,), bucket5))

        tree7 = ts()
        tree7.__setstate__(((bucket7,), bucket7))

        tree11 = ts()
        tree11.__setstate__(((bucket11,), bucket11))

        # Paste together the middle layers.
        tree13 = ts()
        tree13.__setstate__(((tree1, 2, tree3), bucket1))

        tree5711lower = ts()
        tree5711lower.__setstate__(((tree5, 6, tree7, 10, tree11), bucket5))
        tree5711 = ts()
        tree5711.__setstate__(((tree5711lower,), bucket5))

        # One more.
        t = ts()
        t.__setstate__(((tree13, 4, tree5711), bucket1))
        t._check()
        check(t)
        return t, [1, 3, 5, 7, 11]

    def testBasicOps(self):
        t, keys = self._build_degenerate_tree()
        self.assertEqual(len(t), len(keys))
        self.assertEqual(list(t.keys()), keys)

        self.assertTrue(t.has_key(1))
        self.assertTrue(t.has_key(3))
        self.assertTrue(t.has_key(5))
        self.assertTrue(t.has_key(7))
        self.assertTrue(t.has_key(11))
        for i in 0, 2, 4, 6, 8, 9, 10, 12:
            self.assertNotIn(i, t)

    def _checkRanges(self, tree, keys):
        self.assertEqual(len(tree), len(keys))
        sorted_keys = keys[:]
        sorted_keys.sort()
        self.assertEqual(list(tree.keys()), sorted_keys)
        for k in keys:
            self.assertIn(k, tree)
        if keys:
            lokey = sorted_keys[0]
            hikey = sorted_keys[-1]
            self.assertEqual(lokey, tree.minKey())
            self.assertEqual(hikey, tree.maxKey())
        else:
            lokey = hikey = 42

        # Try all range searches.
        for lo in range(lokey - 1, hikey + 2):
            for hi in range(lo - 1, hikey + 2):
                for skipmin in False, True:
                    for skipmax in False, True:
                        wantlo, wanthi = lo, hi
                        if skipmin:
                            wantlo += 1
                        if skipmax:
                            wanthi -= 1
                        want = [k for k in keys if wantlo <= k <= wanthi]
                        got = list(tree.keys(lo, hi, skipmin, skipmax))
                        self.assertEqual(want, got)

    def testRanges(self):
        t, keys = self._build_degenerate_tree()
        self._checkRanges(t, keys)

    def testDeletes(self):
        # Delete keys in all possible orders, checking each tree along
        # the way.

        # This is a tough test.  Previous failure modes included:
        # 1. A variety of assertion failures in _checkRanges.
        # 2. Assorted "Invalid firstbucket pointer" failures at
        #    seemingly random times, coming out of the BTree destructor.
        # 3. Under Python 2.3 CVS, some baffling
        #      RuntimeWarning: tp_compare didn't return -1 or -2 for exception
        #    warnings, possibly due to memory corruption after a BTree
        #    goes insane.
        # On CPython in PURE_PYTHON mode, this is a *slow* test, taking 15+s
        # on a 2015 laptop.
        from BTrees.check import check
        t, keys = self._build_degenerate_tree()
        for oneperm in permutations(keys):
            t, keys = self._build_degenerate_tree()
            for key in oneperm:
                t.remove(key)
                keys.remove(key)
                t._check()
                check(t)
                self._checkRanges(t, keys)
            # We removed all the keys, so the tree should be empty now.
            self.assertEqual(t.__getstate__(), None)

            # A damaged tree may trigger an "invalid firstbucket pointer"
            # failure at the time its destructor is invoked.  Try to force
            # that to happen now, so it doesn't look like a baffling failure
            # at some unrelated line.
            del t   # trigger destructor


LP294788_ids = {}


class ToBeDeleted:
    def __init__(self, id):
        assert isinstance(id, int)  # don't want to store any object ref here
        self.id = id

        LP294788_ids[id] = 1

    def __del__(self):
        LP294788_ids.pop(self.id, None)

    def __le__(self, other):
        return self.id <= other.id

    def __lt__(self, other):
        return self.id < other.id

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return self.id != other.id

    def __gt__(self, other):
        return self.id > other.id

    def __ge__(self, other):
        return self.id >= other.id

    def __hash__(self):
        return hash(self.id)


class TestBugFixes(unittest.TestCase):

    # Collector 1843.  Error returns were effectively ignored in
    # Bucket_rangeSearch(), leading to "delayed" errors, or worse.
    def testFixed1843(self):
        from BTrees.IIBTree import IISet
        t = IISet()
        t.insert(1)
        # This one used to fail to raise the TypeError when it occurred.
        self.assertRaises(TypeError, t.keys, "")
        # This one used to segfault.
        self.assertRaises(TypeError, t.keys, 0, "")

    def test_LP294788(self):
        # https://bugs.launchpad.net/bugs/294788
        # BTree keeps some deleted objects referenced

        # The logic here together with the ToBeDeleted class is that
        # a separate reference dict is populated on object creation
        # and removed in __del__
        # That means what's left in the reference dict is never GC'ed
        # therefore referenced somewhere
        # To simulate real life, some random data is used to exercise the tree
        import gc
        import random

        from BTrees.OOBTree import OOBTree

        t = OOBTree()

        trandom = random.Random('OOBTree')

        global LP294788_ids

        # /// BTree keys are integers, value is an object
        LP294788_ids = {}
        ids = {}
        for i in range(1024):
            if trandom.random() > 0.1 or not ids:
                # add
                id = None
                while id is None or id in ids:
                    id = trandom.randint(0, 1000000)

                ids[id] = 1
                t[id] = ToBeDeleted(id)
            else:
                # del
                keys = list(ids.keys())
                if keys:
                    id = trandom.choice(list(ids.keys()))
                    del t[id]
                    del ids[id]

        ids = ids.keys()
        trandom.shuffle(list(ids))
        for id in ids:
            del t[id]
        ids = None

        # to be on the safe side run a full GC
        gc.collect()

        # print LP294788_ids

        self.assertEqual(len(t), 0)
        self.assertEqual(len(LP294788_ids), 0)
        # \\\

        # /// BTree keys are integers, value is a tuple having an object
        LP294788_ids = {}
        ids = {}
        for i in range(1024):
            if trandom.random() > 0.1 or not ids:
                # add
                id = None
                while id is None or id in ids:
                    id = trandom.randint(0, 1000000)

                ids[id] = 1
                t[id] = (id, ToBeDeleted(id), 'somename')
            else:
                # del
                keys = list(ids.keys())
                if keys:
                    id = trandom.choice(keys)
                    del t[id]
                    del ids[id]

        ids = ids.keys()
        trandom.shuffle(list(ids))
        for id in ids:
            del t[id]
        ids = None

        # to be on the safe side run a full GC
        gc.collect()

        # print LP294788_ids

        self.assertEqual(len(t), 0)
        self.assertEqual(len(LP294788_ids), 0)
        # \\\

        # /// BTree keys are objects, value is an int
        t = OOBTree()
        LP294788_ids = {}
        ids = {}
        for i in range(1024):
            if trandom.random() > 0.1 or not ids:
                # add
                id = None
                while id is None or id in ids:
                    id = ToBeDeleted(trandom.randint(0, 1000000))

                ids[id] = 1
                t[id] = 1
            else:
                # del
                id = trandom.choice(list(ids.keys()))
                del ids[id]
                del t[id]

        ids = ids.keys()
        trandom.shuffle(list(ids))
        for id in ids:
            del t[id]
        # release all refs
        ids = id = None

        # to be on the safe side run a full GC
        gc.collect()

        # print LP294788_ids

        self.assertEqual(len(t), 0)
        self.assertEqual(len(LP294788_ids), 0)

        # /// BTree keys are tuples having objects, value is an int
        t = OOBTree()
        LP294788_ids = {}
        ids = {}
        for i in range(1024):
            if trandom.random() > 0.1 or not ids:
                # add
                id = None
                while id is None or id in ids:
                    id = trandom.randint(0, 1000000)
                    id = (id, ToBeDeleted(id), 'somename')

                ids[id] = 1
                t[id] = 1
            else:
                # del
                id = trandom.choice(list(ids.keys()))
                del ids[id]
                del t[id]

        ids = ids.keys()
        trandom.shuffle(list(ids))
        for id in ids:
            del t[id]
        # release all refs
        ids = id = None

        # to be on the safe side run a full GC
        gc.collect()

        # print LP294788_ids

        self.assertEqual(len(t), 0)
        self.assertEqual(len(LP294788_ids), 0)


# comparison error propagation tests


class DoesntLikeBeingCompared:

    def _cmp(self, other):
        raise ValueError('incomparable')

    __lt__ = __le__ = __eq__ = __ne__ = __ge__ = __gt__ = _cmp


class TestCmpError(unittest.TestCase):

    def testFoo(self):
        from BTrees.OOBTree import OOBTree
        t = OOBTree()
        t['hello world'] = None
        try:
            t[DoesntLikeBeingCompared()] = None
        except ValueError as e:
            self.assertEqual(str(e), 'incomparable')
        else:
            self.fail('incomarable objects should not be allowed into '
                      'the tree')


class FamilyTest(unittest.TestCase):
    def test32(self):
        from zope.interface.verify import verifyObject

        import BTrees
        from BTrees.IOBTree import IOTreeSet
        verifyObject(BTrees.Interfaces.IBTreeFamily, BTrees.family32)
        self.assertEqual(
            BTrees.family32.IO, BTrees.IOBTree)
        self.assertEqual(
            BTrees.family32.OI, BTrees.OIBTree)
        self.assertEqual(
            BTrees.family32.II, BTrees.IIBTree)
        self.assertEqual(
            BTrees.family32.IF, BTrees.IFBTree)
        self.assertEqual(
            BTrees.family32.UO, BTrees.UOBTree)
        self.assertEqual(
            BTrees.family32.OU, BTrees.OUBTree)
        self.assertEqual(
            BTrees.family32.UU, BTrees.UUBTree)
        self.assertEqual(
            BTrees.family32.UF, BTrees.UFBTree)
        self.assertEqual(
            BTrees.family32.OO, BTrees.OOBTree)
        self.assertEqual(
            BTrees.family32.OU, BTrees.OUBTree)
        s = IOTreeSet()
        s.insert(BTrees.family32.maxint)
        self.assertIn(BTrees.family32.maxint, s)
        s = IOTreeSet()
        s.insert(BTrees.family32.minint)
        self.assertIn(BTrees.family32.minint, s)
        s = IOTreeSet()
        # this next bit illustrates an, um, "interesting feature".  If
        # the characteristics change to match the 64 bit version, please
        # feel free to change.
        with self.assertRaises((TypeError, OverflowError)):
            s.insert(BTrees.family32.maxint + 1)

        with self.assertRaises((TypeError, OverflowError)):
            s.insert(BTrees.family32.minint - 1)
        self.check_pickling(BTrees.family32)

    def test64(self):
        from zope.interface.verify import verifyObject

        import BTrees
        from BTrees.LOBTree import LOTreeSet
        verifyObject(BTrees.Interfaces.IBTreeFamily, BTrees.family64)
        self.assertEqual(
            BTrees.family64.IO, BTrees.LOBTree)
        self.assertEqual(
            BTrees.family64.OI, BTrees.OLBTree)
        self.assertEqual(
            BTrees.family64.II, BTrees.LLBTree)
        self.assertEqual(
            BTrees.family64.IF, BTrees.LFBTree)
        self.assertEqual(
            BTrees.family64.UO, BTrees.QOBTree)
        self.assertEqual(
            BTrees.family64.OU, BTrees.OQBTree)
        self.assertEqual(
            BTrees.family64.UU, BTrees.QQBTree)
        self.assertEqual(
            BTrees.family64.UF, BTrees.QFBTree)
        self.assertEqual(
            BTrees.family64.OO, BTrees.OOBTree)
        self.assertEqual(
            BTrees.family64.OU, BTrees.OQBTree)
        s = LOTreeSet()
        s.insert(BTrees.family64.maxint)
        self.assertIn(BTrees.family64.maxint, s)
        s = LOTreeSet()
        s.insert(BTrees.family64.minint)
        self.assertIn(BTrees.family64.minint, s)
        s = LOTreeSet()

        # XXX why oh why do we expect ValueError here, but TypeError in test32?
        with self.assertRaises((TypeError, OverflowError)):
            s.insert(BTrees.family64.maxint + 1)

        with self.assertRaises((TypeError, OverflowError)):
            s.insert(BTrees.family64.minint - 1)

        self.check_pickling(BTrees.family64)

    def check_pickling(self, family):
        # The "family" objects are singletons; they can be pickled and
        # unpickled, and the same instances will always be returned on
        # unpickling, whether from the same unpickler or different
        # unpicklers.
        import pickle
        from io import BytesIO

        s = pickle.dumps((family, family))
        (f1, f2) = pickle.loads(s)
        self.assertIs(f1, family)
        self.assertIs(f2, family)

        # Using a single memo across multiple pickles:
        sio = BytesIO()
        p = pickle.Pickler(sio)
        p.dump(family)
        p.dump([family])
        u = pickle.Unpickler(BytesIO(sio.getvalue()))
        f1 = u.load()
        f2, = u.load()
        self.assertIs(f1, family)
        self.assertIs(f2, family)

        # Using separate memos for each pickle:
        sio = BytesIO()
        p = pickle.Pickler(sio)
        p.dump(family)
        p.clear_memo()
        p.dump([family])
        u = pickle.Unpickler(BytesIO(sio.getvalue()))
        f1 = u.load()
        f2, = u.load()
        self.assertIs(f1, family)
        self.assertIs(f2, family)



##############################################################################
#
# Copyright (c) 2020 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from unittest import TestCase

from ..OOBTree import OOBTree
from .common import ZODBAccess
from .common import _skip_wo_ZODB


BUCKET_SIZE = OOBTree.max_leaf_size


class TestPersistency(ZODBAccess, TestCase):
    @_skip_wo_ZODB
    def test_empty_bucket_persistency(self):
        from transaction import commit
        root = self._getRoot()
        try:
            # tree with 3 buckets (internal implementation details)
            tree = OOBTree(
                {i: i for i in range(3 * BUCKET_SIZE // 2 + 2)})
            root["tree"] = tree
            commit()
            # almost clear the second bucket keeping the last element
            for i in range(BUCKET_SIZE // 2, BUCKET_SIZE - 1):
                del tree[i]
            commit()
            del tree[BUCKET_SIZE - 1]  # remove the last element
            commit()
            tree._check()
            tree._p_deactivate()
            tree._check()  # fails in case of bad persistency
        finally:
            self._closeRoot(root)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from zope.interface import Attribute
from zope.interface import Interface
from zope.interface.common.collections import IMapping
from zope.interface.common.collections import ISized
from zope.interface.common.sequence import IMinimalSequence


class ICollection(Interface):
    """
    A collection of zero or more objects.

    In a boolean context, objects implementing this interface are
    `True` if the collection is non-empty, and `False` if the
    collection is empty.
    """

    def clear():
        """Remove all of the items from the collection."""


# Backwards compatibility alias. To be removed in 5.0.
# Docs deprecated only in docs/api.rst.
IReadSequence = IMinimalSequence


class IKeyed(ICollection):

    def has_key(key):
        """Check whether the object has an item with the given key.

        Return a true value if the key is present, else a false value.
        """

    def keys(min=None, max=None, excludemin=False, excludemax=False):
        """
        Return an :mod:`IMinimalSequence <zope.interface.common.sequence>`
        containing the keys in the collection.

        The type of the ``IMinimalSequence`` is not specified. It
        could be a `list` or a `tuple` or some other type.

        All arguments are optional, and may be specified as keyword
        arguments, or by position.

        If a *min* is specified, then output is constrained to keys
        greater than or equal to the given min, and, if *excludemin*
        is specified and true, is further constrained to keys strictly
        greater than *min*. A *min* value of `None` is ignored. If
        *min* is `None` or not specified, and *excludemin* is true,
        the smallest key is excluded.

        If a *max* is specified, then output is constrained to keys
        less than or equal to the given *max*, and, if *excludemax* is
        specified and true, is further constrained to keys strictly
        less than *max*. A *max* value of `None` is ignored. If *max*
        is `None` or not specified, and *excludemax* is true, the
        largest key is excluded.
        """

    def maxKey(key=None):
        """Return the maximum key.

        If a key argument if provided and not None, return the largest key
        that is less than or equal to the argument.  Raise an exception if
        no such key exists.
        """

    def minKey(key=None):
        """Return the minimum key.

        If a key argument if provided and not None, return the smallest key
        that is greater than or equal to the argument.  Raise an exception
        if no such key exists.
        """


class ISetMutable(IKeyed):

    def insert(key):
        """Add the key (value) to the set.

        If the key was already in the set, return 0, otherwise return 1.
        """

    def remove(key):
        """Remove the key from the set.

        Raises :class:`KeyError` if key is not in the set.
        """

    def update(seq):
        """Add the items from the given sequence to the set."""

    def __and__(other):
        """
        Shortcut for :meth:`~BTrees.Interfaces.IMerge.intersection`

        .. versionadded:: 4.8.0
        """

    def __iand__(other):
        """
        As for :meth:`set.intersection_update`: Update this object,
        keeping only elements found in both it and other.

        .. versionadded:: 4.8.0
        """

    def __or__(other):
        """
        Shortcut for :meth:`~BTrees.Interfaces.IMerge.union`

        .. versionadded:: 4.8.0
        """

    def __ior__(other):
        """
        As for :meth:`set.update`: Update this object, adding
        elements from *other*.

        .. versionadded:: 4.8.0
        """

    def __sub__(other):
        """
        Shortcut for :meth:`~BTrees.Interfaces.IMerge.difference`

        .. versionadded:: 4.8.0
        """

    def __isub__(other):
        """
        As for :meth:`set.difference_update`: Update this object,
        removing elements found in *other*.

        .. versionadded:: 4.8.0
        """

    def isdisjoint(other):
        """
        As for :meth:`set.isdisjoint`: Return True if the set has no
        elements in common with other.

        .. versionadded:: 4.8.0
        """

    def discard(key):
        """
        As for :meth:`set.discard`: Remove the *key* from the set,
        but only if it is present.

        .. versionadded:: 4.8.0
        """

    def pop():
        """
        As for :meth:`set.pop`: Remove and return an arbitrary element;
        raise :exc:`KeyError` if the object is empty.

        .. versionadded:: 4.8.0
        """

    def __ixor__(other):
        """
        As for :meth:`set.symmetric_difference_update`: Update this object,
        keeping only elements found in either set but not in both.

        .. versionadded:: 4.8.0
        """


class IKeySequence(IKeyed, ISized):

    def __getitem__(index):
        """Return the key in the given index position.

        This allows iteration with for loops and use in functions,
        like map and list, that read sequences.
        """


class ISet(ISetMutable, IKeySequence):
    """
    A set of unique items stored in a single persistent object.
    """


class ITreeSet(ISetMutable):
    """
    A set of unique items stored in a tree of persistent objects.
    """


class IMinimalDictionary(IKeyed, IMapping):
    """
    Mapping operations.

    .. versionchanged:: 4.8.0
       Now extends :class:`zope.interface.common.collections.IMapping`.
    """

    def get(key, default):
        """Get the value associated with the given key.

        Return the default if :meth:`~BTrees.Interfaces.IKeyed.has_key` is
        false with the given key.
        """

    def __getitem__(key):
        """Get the value associated with the given key.

        Raise :class:`KeyError` if :meth:`~BTrees.Interfaces.IKeyed.has_key`
        is false with the given key.
        """

    def __setitem__(key, value):
        """Set the value associated with the given key."""

    def __delitem__(key):
        """Delete the value associated with the given key.

        Raise class:`KeyError` if :meth:`~BTrees.Interfaces.IKeyed.has_key` is
        false with the given key.
        """

    def values(min=None, max=None, excludemin=False, excludemax=False):
        """ Return a minimal sequence containing the values in the collection.

        Return value is an
        :mod:`IMinimalSequence
        <zope.interface.common.sequence.IMinimalSequence>`.

        The type of the ``IMinimalSequence`` is not specified. It
        could be a `list` or a `tuple` or some other type.

        All arguments are optional, and may be specified as keyword
        arguments, or by position.

        If a *min* is specified, then output is constrained to values
        whose keys are greater than or equal to the given *min*, and, if
        *excludemin* is specified and true, is further constrained to
        values whose keys are strictly greater than *min*. A *min* value
        of `None` is ignored. If *min* is `None` or not specified, and
        *excludemin* is true, the value corresponding to the smallest
        key is excluded.

        If a *max* is specified, then output is constrained to values
        whose keys are less than or equal to the given *max*, and, if
        *excludemax* is specified and true, is further constrained to
        values whose keys are strictly less than *max*. A *max* value of
        `None` is ignored. If *max* is `None` or not specified, and
        *excludemax* is true, the value corresponding to the largest key
        is excluded.
        """

    def items(min=None, max=None, excludemin=False, excludemax=False):
        """
        Return an ``IMinimalSequence`` containing the items in the
        collection.

        An item is a 2-tuple, a ``(key, value)`` pair.

        The type of the ``IMinimalSequence`` is not specified. It
        could be a `list` or a `tuple` or some other type.

        All arguments are optional, and may be specified as keyword
        arguments, or by position.

        If a *min* is specified, then output is constrained to items
        whose keys are greater than or equal to the given *min*, and,
        if *excludemin* is specified and true, is further constrained
        to items whose keys are strictly greater than *min*. A *min*
        value of `None` is ignored. If *min* is `None` or not
        specified, and *excludemin* is true, the item with the
        smallest key is excluded.

        If a *max* is specified, then output is constrained to items
        whose keys are less than or equal to the given *max*, and, if
        *excludemax is specified and true, is further constrained to
        items whose keys are strictly less than *max*. A *max* value
        of `None` is ignored. If *max* is `None` or not specified, and
        *excludemax* is true, the item with the largest key is
        excluded.
        """


class IDictionaryIsh(IMinimalDictionary):

    def update(collection):
        """Add the items from the given collection object to the collection.

        The input collection must be a sequence of (key, value) 2-tuples,
        or an object with an 'items' method that returns a sequence of
        (key, value) pairs.
        """

    def byValue(minValue):
        """Return a sequence of (value, key) pairs, sorted by value.

        Values < minValue are omitted and other values are "normalized" by
        the minimum value.  This normalization may be a noop, but, for
        integer values, the normalization is division.
        """

    def setdefault(key, d):
        """D.setdefault(k, d) -> D.get(k, d), also set D[k]=d if k not in D.

        Return the value like
        :meth:`~BTrees.Interfaces.IMinimalDictionary.get` except that if key
        is missing, d is both returned and inserted into the dictionary as the
        value of k.

        Note that, unlike as for Python's :meth:`dict.setdefault`, d is not
        optional.  Python defaults d to None, but that doesn't make sense
        for mappings that can't have None as a value (for example, an
        :class:`~BTrees.IIBTree.IIBTree` can have only integers as values).
        """

    def pop(key, d):
        """D.pop(k[, d]) -> v, remove key and return the corresponding value.

        If key is not found, d is returned if given, otherwise
        :class:`KeyError` is raised.
        """

    def popitem():
        """
        D.popitem() -> (k, v), remove and return some (key, value) pair
        as a 2-tuple; but raise KeyError if D is empty.

        .. versionadded:: 4.8.0
        """


class IBTree(IDictionaryIsh):

    def insert(key, value):
        """Insert a key and value into the collection.

        If the key was already in the collection, then there is no
        change and 0 is returned.

        If the key was not already in the collection, then the item is
        added and 1 is returned.

        This method is here to allow one to generate random keys and
        to insert and test whether the key was there in one operation.

        A standard idiom for generating new keys will be::

          key = generate_key()
          while not t.insert(key, value):
              key=generate_key()
        """

    def __and__(other):
        """Shortcut for :meth:`~BTrees.Interfaces.IMerge.intersection`"""

    def __or__(other):
        """Shortcut for :meth:`~BTrees.Interfaces.IMerge.union`"""

    def __sub__(other):
        """Shortcut for :meth:`~BTrees.Interfaces.IMerge.difference`"""


class IMerge(Interface):
    """Object with methods for merging sets, buckets, and trees.

    These methods are supplied in modules that define collection
    classes with particular key and value types. The operations apply
    only to collections from the same module.  For example, the
    :meth:`BTrees.IIBTree.IIBTree.union` can only be used with
    :class:`~BTrees.IIBTree.IIBTree`, :class:`~BTrees.IIBTree.IIBucket`,
    :class:`~BTrees.IIBTree.IISet`, and :class:`~BTrees.IIBTree.IITreeSet`.

    The number protocols methods ``__and__``, ``__or__`` and ``__sub__`` are
    provided by all the data structures. They are shortcuts for
    :meth:`~BTrees.Interfaces.IMerge.intersection`,
    :meth:`~BTrees.Interfaces.IMerge.union` and
    :meth:`~BTrees.Interfaces.IMerge.difference`.

    The implementing module has a value type. The
    :class:`~BTrees.IOBTree.IOBTree` and :class:`~BTrees.OOBTree.OOBTree`
    modules have object value type. The :class:`~BTrees.IIBTree.IIBTree` and
    :class:`~BTrees.OIBTree.OIBTree` modules have integer value types. Other
    modules may be defined in the future that have other value types.

    The individual types are classified into set (Set and TreeSet) and
    mapping (Bucket and BTree) types.
    """

    def difference(c1, c2):
        """Return the keys or items in c1 for which there is no key in c2.

        If c1 is None, then None is returned.  If c2 is None, then c1
        is returned.

        If neither c1 nor c2 is None, the output is a Set if c1 is a Set or
        TreeSet, and is a Bucket if c1 is a Bucket or BTree.

        While *c1* must be one of those types, *c2* can be any Python iterable
        returning the correct types of objects.

        .. versionchanged:: 4.8.0
           Add support for *c2* to be an arbitrary iterable.
        """

    def union(c1, c2):
        """Compute the Union of c1 and c2.

        If c1 is None, then c2 is returned, otherwise, if c2 is None,
        then c1 is returned.

        The output is a Set containing keys from the input
        collections.

        *c1* and *c2* can be any Python iterables returning the
        correct type of objects.

        .. versionchanged:: 4.8.0
           Add support for arbitrary iterables.
        """

    def intersection(c1, c2):
        """Compute the intersection of c1 and c2.

        If c1 is None, then c2 is returned, otherwise, if c2 is None,
        then c1 is returned.

        The output is a Set containing matching keys from the input
        collections.

        *c1* and *c2* can be any Python iterables returning the
        correct type of objects.

        .. versionchanged:: 4.8.0
           Add support for arbitrary iterables.
        """


class IBTreeModule(Interface):
    """These are available in all modules (IOBTree, OIBTree, OOBTree, IIBTree,
    IFBTree, LFBTree, LOBTree, OLBTree, and LLBTree).
    """

    BTree = Attribute(
        """The IBTree for this module.

        Also available as [prefix]BTree, as in IOBTree.""")

    Bucket = Attribute(
        """The leaf-node data buckets used by the BTree.

        (IBucket is not currently defined in this file, but is essentially
        IDictionaryIsh, with the exception of __nonzero__, as of this
        writing.)

        Also available as [prefix]Bucket, as in IOBucket.""")

    TreeSet = Attribute(
        """The ITreeSet for this module.

        Also available as [prefix]TreeSet, as in IOTreeSet.""")

    Set = Attribute(
        """The ISet for this module: the leaf-node data buckets used by the
        TreeSet.

        Also available as [prefix]BTree, as in IOSet.""")


class IIMerge(IMerge):
    """Merge collections with integer value type.

    A primary intent is to support operations with no or integer
    values, which are used as "scores" to rate indiviual keys. That
    is, in this context, a BTree or Bucket is viewed as a set with
    scored keys, using integer scores.
    """

    def weightedUnion(c1, c2, weight1=1, weight2=1):
        """Compute the weighted union of c1 and c2.

        If c1 and c2 are None, the output is (0, None).

        If c1 is None and c2 is not None, the output is (weight2, c2).

        If c1 is not None and c2 is None, the output is (weight1, c1).

        Else, and hereafter, c1 is not None and c2 is not None.

        If c1 and c2 are both sets, the output is 1 and the (unweighted)
        union of the sets.

        Else the output is 1 and a Bucket whose keys are the union of c1 and
        c2's keys, and whose values are::

          v1*weight1 + v2*weight2

          where:

            v1 is 0        if the key is not in c1
                  1        if the key is in c1 and c1 is a set
                  c1[key]  if the key is in c1 and c1 is a mapping

            v2 is 0        if the key is not in c2
                  1        if the key is in c2 and c2 is a set
                  c2[key]  if the key is in c2 and c2 is a mapping

        Note that c1 and c2 must be collections.
        """

    def weightedIntersection(c1, c2, weight1=1, weight2=1):
        """Compute the weighted intersection of c1 and c2.

        If c1 and c2 are None, the output is (0, None).

        If c1 is None and c2 is not None, the output is (weight2, c2).

        If c1 is not None and c2 is None, the output is (weight1, c1).

        Else, and hereafter, c1 is not None and c2 is not None.

        If c1 and c2 are both sets, the output is the sum of the weights
        and the (unweighted) intersection of the sets.

        Else the output is 1 and a Bucket whose keys are the intersection of
        c1 and c2's keys, and whose values are::

          v1*weight1 + v2*weight2

          where:

            v1 is 1        if c1 is a set
                  c1[key]  if c1 is a mapping

            v2 is 1        if c2 is a set
                  c2[key]  if c2 is a mapping

        Note that c1 and c2 must be collections.
        """


class IMergeIntegerKey(IMerge):
    """:class:`~BTrees.Interfaces.IMerge`-able objects with integer keys.

    Concretely, this means the types in :class:`~BTree.IOBTree.IOBTree` and
    :class:`~BTrees.IIBTree.IIBTree`.
    """

    def multiunion(seq):
        """Return union of (zero or more) integer sets, as an integer set.

        seq is a sequence of objects each convertible to an integer set.
        These objects are convertible to an integer set:

        + An integer, which is added to the union.

        + A Set or TreeSet from the same module (for example, an
          :class:`BTrees.IIBTree.TreeSet` for
          :meth:`BTrees.IIBTree.multiunion`).  The elements of the set are
          added to the union.

        + A Bucket or BTree from the same module (for example, an
          :class:`BTrees.IOBTree.IOBTree` for
          :meth:`BTrees.IOBTree.multiunion`).  The keys of the mapping are
          added to the union.

        + Any iterable Python object that iterates across integers. This
          will be slower than the above types.

        The union is returned as a Set from the same module (for example,
        :meth:`BTrees.IIBTree.multiunion` returns an
        :class:`BTrees.IIBTree.IISet`).

        The point to this method is that it can run much faster than doing a
        sequence of two-input :meth:`~BTrees.Interfaces.IMerge.union` calls.
        Under the covers, all the integers in all the inputs are sorted via a
        single linear-time radix sort, then duplicates are removed in a second
        linear-time pass.

        .. versionchanged:: 4.8.0
           Add support for arbitrary iterables of integers.
        """


class IBTreeFamily(Interface):
    """the 64-bit or 32-bit family"""
    IF = Attribute('The IIntegerFloatBTreeModule for this family')
    II = Attribute('The IIntegerIntegerBTreeModule for this family')
    IO = Attribute('The IIntegerObjectBTreeModule for this family')
    IU = Attribute('The IIntegerUnsignedBTreeModule for this family')

    UF = Attribute('The IUnsignedFloatBTreeModule for this family')
    UI = Attribute('The IUnsignedIntegerBTreeModule for this family')
    UO = Attribute('The IUnsignedObjectBTreeModule for this family')
    UU = Attribute('The IUnsignedUnsignedBTreeModule for this family')

    OI = Attribute('The IObjectIntegerBTreeModule for this family')
    OO = Attribute('The IObjectObjectBTreeModule for this family')
    OU = Attribute('The IObjectUnsignedBTreeModule for this family')

    maxint = Attribute('The maximum signed integer storable in this family')
    maxuint = Attribute('The maximum unsigned integer storable in this family')
    minint = Attribute('The minimum signed integer storable in this family')


class _IMergeBTreeModule(IBTreeModule, IMerge):
    family = Attribute('The IBTreeFamily of this module')


class IIntegerObjectBTreeModule(_IMergeBTreeModule):
    """Keys, or set values, are signed ints; values are objects.

    Describes IOBTree and LOBTree.
    """


class IUnsignedObjectBTreeModule(_IMergeBTreeModule):
    """Keys, or set values, are unsigned ints; values are objects.

    Describes UOBTree and QOBTree.
    """


class IObjectIntegerBTreeModule(_IMergeBTreeModule):
    """Keys, or set values, are objects; values are signed ints.

    Object keys (and set values) must sort reliably (for instance, *not* on
    object id)!  Homogenous key types recommended.

    Describes OIBTree and OLBTree.
    """


class IObjectUnsignedBTreeModule(_IMergeBTreeModule):
    """Keys, or set values, are objects; values are signed ints.

    Object keys (and set values) must sort reliably (for instance, *not* on
    object id)!  Homogenous key types recommended.

    Describes OUBTree and OQBTree.
    """


class IIntegerIntegerBTreeModule(_IMergeBTreeModule, IMergeIntegerKey):
    """Keys, or set values, are signed integers; values are signed integers.

    Describes IIBTree and LLBTree
    """


class IUnsignedUnsignedBTreeModule(_IMergeBTreeModule, IMergeIntegerKey):
    """Keys, or set values, are unsigned ints; values are unsigned ints.

    Describes UUBTree and QQBTree
    """


class IUnsignedIntegerBTreeModule(_IMergeBTreeModule, IMergeIntegerKey):
    """Keys, or set values, are unsigned ints; values are signed ints.

    Describes UIBTree and QLBTree
    """


class IIntegerUnsignedBTreeModule(_IMergeBTreeModule, IMergeIntegerKey):
    """Keys, or set values, are signed ints; values are unsigned ints.

    Describes IUBTree and LQBTree
    """


class IObjectObjectBTreeModule(IBTreeModule, IMerge):
    """Keys, or set values, are objects; values are also objects.

    Object keys (and set values) must sort reliably (for instance, *not* on
    object id)!  Homogenous key types recommended.

    Note that there's no ``family`` attribute; all families include
    the OO flavor of BTrees.

    Describes OOBTree
    """


class IIntegerFloatBTreeModule(_IMergeBTreeModule):
    """Keys, or set values, are signed ints; values are floats.

    Describes IFBTree and LFBTree
    """


class IUnsignedFloatBTreeModule(_IMergeBTreeModule):
    """Keys, or set values, are unsigned ints; values are floats.

    Describes UFBTree and QFBTree
    """


try:
    from ZODB.POSException import BTreesConflictError
except ModuleNotFoundError:
    class BTreesConflictError(ValueError):
        @property
        def reason(self):
            return self.args[-1]

###############################################################
# IMPORTANT NOTE
#
# Getting the length of a BTree, TreeSet, or output of keys,
# values, or items of same is expensive. If you need to get the
# length, you need to maintain this separately.
#
# Eventually, I need to express this through the interfaces.
#
################################################################



##############################################################################
#
# Copyright 2012 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest

from BTrees import _datatypes


to_ob = _datatypes.Any()
to_int = _datatypes.I()
to_float = _datatypes.F()
to_long = _datatypes.L()
to_2_bytes = _datatypes.f()
to_6_bytes = _datatypes.s()


class TestDatatypes(unittest.TestCase):
    def test_to_ob(self):
        for thing in "abc", 0, 1.3, (), frozenset((1, 2)), object():
            self.assertIs(to_ob(thing), thing)

    def test_to_int_w_int(self):
        self.assertEqual(to_int(3), 3)

    def test_to_int_w_overflow(self):
        self.assertRaises(TypeError, to_int, 2**64)

    def test_to_int_w_invalid(self):
        self.assertRaises(TypeError, to_int, ())

    def test_to_float_w_float(self):
        self.assertEqual(to_float(3.14159), 3.14159)

    def test_to_float_w_int(self):
        self.assertEqual(to_float(3), 3.0)

    def test_to_float_w_invalid(self):
        self.assertRaises(TypeError, to_float, ())

    def test_to_long_w_int(self):
        self.assertEqual(to_long(3), 3)

    def test_to_long_w_overflow(self):
        self.assertRaises(TypeError, to_long, 2**64)

    def test_to_long_w_invalid(self):
        self.assertRaises(TypeError, to_long, ())

    def test_to_2_bytes_w_ok(self):
        self.assertEqual(to_2_bytes(b'ab'), b'ab')

    def test_to_2_bytes_w_invalid_length(self):
        self.assertRaises(TypeError, to_2_bytes, b'a')
        self.assertRaises(TypeError, to_2_bytes, b'abcd')

    def test_to_6_bytes_w_ok(self):
        self.assertEqual(to_6_bytes(b'abcdef'), b'abcdef')

    def test_to_6_bytes_w_invalid_length(self):
        self.assertRaises(TypeError, to_6_bytes, b'a')
        self.assertRaises(TypeError, to_6_bytes, b'abcd')

    def test_coerce_to_6_bytes(self):
        # correct input is passed through
        self.assertEqual(to_6_bytes.coerce(b'abcdef'), b'abcdef')

        # small positive integers are converted
        self.assertEqual(to_6_bytes.coerce(1), b'\x00\x00\x00\x00\x00\x01')

        # negative values are disallowed
        self.assertRaises(TypeError, to_6_bytes.coerce, -1)

        # values outside the bigger than 64-bits are disallowed
        self.assertRaises(TypeError, to_6_bytes.coerce, 2 ** 64 + 1)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import unittest

from BTrees.OOBTree import OOBTree
from BTrees.OOBTree import OOBucket


class B(OOBucket):
    pass


class T(OOBTree):
    _bucket_type = B
    max_leaf_size = 2
    max_internal_size = 3


class S(T):
    pass


class SubclassTest(unittest.TestCase):

    def testSubclass(self):
        # test that a subclass that defines _bucket_type gets buckets
        # of that type
        t = T()
        t[0] = 0
        self.assertIs(t._firstbucket.__class__, B)

    def testCustomNodeSizes(self, TreeKind=S, BucketKind=B):
        # We override btree and bucket split sizes in BTree subclasses.
        t = TreeKind()
        for i in range(8):
            t[i] = i

        state = t.__getstate__()[0]
        self.assertEqual(len(state), 5)
        sub = state[0]
        # __class__ is a property in the Python implementation, and
        # if the C extension is available it returns the C version.
        self.assertIsInstance(sub, TreeKind)
        sub = sub.__getstate__()[0]
        self.assertEqual(len(sub), 5)
        sub = sub[0]
        self.assertIsInstance(sub, BucketKind)
        self.assertEqual(len(sub), 1)

    def _checkReplaceNodeSizes(self, TreeKind, BucketKind):
        # We can also change the node sizes globally.
        orig_leaf = TreeKind.max_leaf_size
        orig_internal = TreeKind.max_internal_size
        TreeKind.max_leaf_size = T.max_leaf_size
        TreeKind.max_internal_size = T.max_internal_size
        try:
            self.testCustomNodeSizes(TreeKind, BucketKind)
        finally:
            TreeKind.max_leaf_size = orig_leaf
            TreeKind.max_internal_size = orig_internal

    def testReplaceNodeSizesNative(self):
        self._checkReplaceNodeSizes(OOBTree, OOBucket)

    def testReplaceNodeSizesPython(self):
        from BTrees.OOBTree import OOBTreePy
        from BTrees.OOBTree import OOBucketPy
        self._checkReplaceNodeSizes(OOBTreePy, OOBucketPy)
